/* $Id: adjobj.cpp,v 1.11 2007/12/05 15:03:46 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <cassert>
#include "adjobj.h"

namespace DEP
{

// constructor

AdjObject::AdjObject(size_t sz) : _id(), _inuse(false),
_approx_flag(false),	// no approximations exist
_index(-1),				// invalid index in normal equation matrix
_size(sz),				// number of parameter when known
_upari(0)				// not available, created by pure virtual method set_upari
{
	// empty
}


// destructor

AdjObject::~AdjObject()
{
	delete[] _upari;
}


// get string of parameter status

string AdjObject::state_str(size_t i) const
{
	ostringstream oss;

	switch (state(i))
		{
		case unk: oss << "unk"; break;
		case obs: oss << "obs"; break;
		case eli: oss << "eli"; break;
		default:
			assert(false);
		}

	return oss.str();
}


// count number of unknowns by state

size_t AdjObject::num(u_state state) const
{
	assert(_upari != 0);

	size_t n = 0;
	for (size_t i = 0; i < _size; i++)
		{
		if (_upari[i].state == state) n++;
		}

	return n;
}


// print unknown parameters

void AdjObject::print_info(ostream& os) const
{
	Form str5; str5.left().width(5);
	Form str8; str8.left().width(8);
	Form str10; str10.width(10);
	Form int5; int5.width(5);
	Form sci9(2); sci9.scientific().width(9);

	for (size_t i = 0; i < _size; i++)
		os << "  " << int5(_index+i) << "   " << str8(uname(i)) << "   "
			<< str10(_id.c_str()) << "   " << str5(state_str(i)) << "   "
			<< sci9(pii(i)) << endl;
	os << endl;
}


// stochastic weighting of the unknowns to the normal matrix

void AdjObject::stoch_weighting(Matrix<double>& atpa,
Vector<double>& atpl, double& ltpl) const
{
	for (size_t i = 0, k = _index; i < _size; i++, k++)
		{
		if (state(i) == obs || state(i) == eli)
			{

			double p = pii(i);
			//cout << "p= " << p << endl;
			atpa(k,k) += p;

			double l = ival(i) - par(i); // observed - approximated
			atpl(k) += p * l;
			ltpl += p * l * l;
			}
		}
}


// update unknown parameters
// return the number of updated unknowns

size_t AdjObject::update_unknowns(const Vector<double>& dx)
{
	assert(_upari != 0);

	size_t n = 0;

	for (size_t i = 0, k = _index; i < _size; i++, k++)
		{
		//assert(_upari[i].par != 0);
		//if (_upari[i].state != eli)
		*(_upari[i].par) += dx(k);
		n++;
		}

	return n;
}


// update std. devs. of parameters
// return the number of updated unknowns

size_t AdjObject::update_sdev(double sigma0, const Matrix<double>& qxx)
{
	assert(_upari != 0);

	size_t n = 0;

	for (size_t i = 0, k = _index; i < _size; i++, k++)
		{
		//assert(_upari[i].sdev != 0);
		if (_upari[i].state != eli)
			{
			*(_upari[i].sdev) = sqrt(qxx(k,k)) * sigma0;
			n++;
			}
		}

	return n;
}


// get adjustment object class name

string AdjObject::class_str() const
{
	ostringstream oss;
	oss << "undef.";
	return oss.str();
}


// get type name

string AdjObject::type_str() const
{
	ostringstream oss;
	oss << "";
	return oss.str();
}


// get key name

string AdjObject::key_str() const
{
	ostringstream oss;
	oss << "undef";
	return oss.str();
}


// get name of unknown parameter

string AdjObject::uname(size_t i) const
{
	assert(i < _size);

	ostringstream oss;
	oss << "x[" << i << "]";
	return oss.str();
}


// print changes of unknown parameters

void AdjObject::print_changes(ostream& os, const Vector<double>& dx) const
{
	os << "ID       :  " << _id << endl;
	//os << "Class :  " << class_str() << endl;
	string str = type_str();
	if (!str.empty()) os << "Subtype  :  " << str << endl;
	os << "Index    :  " << _index << endl;
	os << endl;
	
	os << "  Param.     State"
		"       Approx.         Change       Adjusted\n\n";

	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form sci12(3); sci12.scientific().width(12);
	
	int k = _index;
	for (size_t i = 0; i < _size; i++, k++)
		os << "  " << str8(uname(i)) << "   " << str4(state_str(i))
			<< "   " << sci12(par(i)) << "   " << sci12(dx(k))
			<< "   " << sci12(par(i) + dx(k)) << endl;
	os << endl;
}


// print unknown parameters

void AdjObject::print_unknowns(ostream& os) const
{
	os << "ID       :  " << _id << endl;
	//os << "Class :  " << class_str() << endl;
	string str = type_str();
	if (!str.empty()) os << "Subtype  :  " << str << endl;
	os << "Index    :  " << _index << endl;
	//os << "Approx.  :  " << (_approx_flag ? "yes" : "no") << endl;
	os << endl;

	os << "  Param.    State      Initial      Adjusted        Change      "
		"Std.dev.\n\n";
	
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form sci12(3); sci12.scientific().width(12);

	for (size_t i = 0; i < _size; i++)
		{
		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< sci12(ival(i)) << "  "
			<< sci12(par(i)) << "  "
			<< sci12(change(i)) << "  "
			<< sci12(sdev(i)) << endl;
		}
	os << endl;
}


// print precision of unknowns

void AdjObject::print_precision(ostream& os) const
{
	os << "ID       :  " << _id << endl;
	//os << "Class :  " << class_str() << endl;
	string str = type_str();
	if (!str.empty()) os << "Subtype  :  " << str << endl;
	os << "Index    :  " << _index << endl;
	os << endl;

	os << "  Param.    State     Initial     Adjusted       Change     "
		"Std.dev.   Ratio\n\n";
	
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form sci11(3); sci11.scientific().width(11);
	Form fix3(1); fix3.fixed().width(6);

	for (size_t i = 0; i < _size; i++)
		{
		double t_abs = ratio(i); // test value

		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< sci11(ival(i)) << "  "
			<< sci11(par(i)) << "  "
			<< sci11(change(i)) << "  "
			<< sci11(sdev(i)) << "  " << fix3(t_abs);
			
		// test on significance
		if (state(i) == eli)
			os << "  is eliminated";
		else
			{
			if (chk_sign(t_abs))
				os << "  is significant";
			else
				os << "  not significant";
			}
		os << endl;
		}
	os << endl;
}


// check on convergence
/*
int AdjObject::check_convergence(const Vector<double>& dx) const
{
	for (int i = 0, k = _index; i < _size; i++, k++)
		{
		if (fabs(dx(k)) > lspars.conv_gen()) return i;
		}
	return -1;
}
*/


// sum of absolute changes

void AdjObject::sum_changes(const Vector<double>& dx,
double& sum, size_t& num) const
{
	int k = _index;
	for (size_t i = 0; i < _size; i++, k++)
		{
		if (state(i) != eli) sum += fabs(dx(k)), num++;
		}
}


// create table

bool AdjObject::create()
{
	assert(_upari == 0);

	_upari = new Upari[_size];
	if (_upari == 0) return false;

	return true;
}


// set table entry

void AdjObject::set(size_t i, double* par, double* sdev)
{
	assert(i < _size);

	if (is_obs(*sdev))
		{
		// observed parameter (real std. devs.)
		_upari[i].state = obs;
		_upari[i].pii = lspars._sigma0 / *sdev;
		_upari[i].pii *= _upari[i].pii;
		}
	else if (is_eli(*sdev))
		{
		// eliminated parameter
		_upari[i].state = eli;
		_upari[i].pii = lspars._unksup_wt;
		}
	else
		{
		// unknown parameter
		_upari[i].state = unk;
		_upari[i].pii = 0.;
		}

	_upari[i].ival = *par;	// save initial value
	_upari[i].par = par;	// pointer to parameter
	_upari[i].sdev = sdev;	// pointer to std. dev.
}


// get state
/*
u_state AdjObject::state(size_t i) const
{
	assert(i < _size);

	double sdev = *(_upari[i].sdev);
	u_state state = unk; // unknown parameter

	if (is_obs(sdev))
		{
		// observed parameter (real std. devs.)
		state = obs;
		}
	else if (is_eli(sdev))
		{
		// eliminated parameter
		state = eli;
		}

	return state;
}
*/

// get weight
/*
double AdjObject::pii(size_t i) const
{
	assert(i < _size);

	double sdev = *(_upari[i].sdev);
	double pii = 0.; // unknown parameter

	if (is_obs(sdev))
		{
		// observed parameter (real std. devs.)
		pii = lspars._sigma0 / sdev;
		pii *= pii;
		}
	else if (is_eli(sdev))
		{
		// eliminated parameter
		pii = lspars._unksup_wt;
		}

	return pii;
}
*/

// get c-style string of parameter status from std. dev.

string AdjObject::state_str(double s) const
{
	ostringstream oss;

	if (s < 0.)
	{
		cout << "ERROR: Invalid standard deviation: " << s << endl;
		assert(false);
	}
	else if (s < lspars._smin)
		oss << "eli"; // eliminated parameter	
	else if (s < lspars._smax)
		oss << "obs"; // observed parameter (real std. devs.)
	else
		oss << "unk"; // unknown parameter

	return oss.str();
}


// print standard deviation

void AdjObject::print_sdev(ostream& os, const Form& fix, double s, double sc) const
{
	if (is_rsdev(s))
		{
		// scaling and print in given format
		os << fix(s * sc);
		}
	else
		{
		// no scaling and print in given scientific format
		Form sci(1); sci.scientific().width(12);
		os << sci(s);
		}
}


// non-member functions ...

// correlations, output in table form

void corr_table(ostream& os, const Matrix<double>& qxx,
const AdjObject& ao1, const AdjObject& ao2, double lim)
{
	Form str8; str8.left().width(8);
	Form str16; str16.left().width(16);
	Form int5; int5.width(5);
	Form fix6(3); fix6.fixed().width(6);

	int k = ao1._index;
	for (size_t i = 0; i < ao1._size; i++, k++)
		{
		int l = ao2._index;
		for (size_t j = 0; j < ao2._size; j++, l++)
			{
			if (k == l) continue; // always r = 1!!!
			double r = qxx(k,l) / sqrt(qxx(k,k)) / sqrt(qxx(l,l));
			
			if (fabs(r) > lim) 
				{
				os << "  " << int5(k) << "  " << int5(l)
					<< "    " << str8(ao1.uname(i))
					<< "  " << str16(ao1.get_id())
					<< "  " << str8(ao2.uname(j))
					<< "  " << str16(ao2.get_id())
					<< "  " << fix6(r) << endl;
				}
			}
		}
}


// correlations, output in matrix form

void corr_matrix(ostream& os, const Matrix<double>& qxx,
const AdjObject& ao1, const AdjObject& ao2, double lim)
{
	//os << "Correlation coeffs. between #" << ao1._id
	//	<< " and #" << ao2._id << "\n\n";

	Form str; str.right().width(8);
	Form fix(3); fix.fixed().width(8);

	os << "         ";
	for (size_t j = 0; j < ao2._size; j++)
		os << " " << str(ao2.uname(j));
	os << "\n\n";

	str.left();

	int k = ao1._index;
	for (size_t i = 0; i < ao1._size; i++, k++)
		{
		os << str(ao1.uname(i)) << "  ";
		int l = ao2._index;
		for (size_t j = 0; j < ao2._size; j++, l++)
			{
			double r = qxx(k,l) / sqrt(qxx(k,k)) / sqrt(qxx(l,l));
			os << fix(r);
			if (k != l && fabs(r) > lim)
				os << '*';
			else
				os << ' ';
			}
		os << endl;
		}
	os << endl;
}


// print matrix of covariance

void covar_matrix(ostream& os, const DMatrix& qxx,
const AdjObject& ao1, const AdjObject& ao2, double sigma0)
{
	Form str8; str8.right().width(8);
	Form str11; str11.right().width(11);
	Form sci(3); sci.scientific().width(11);

	os << "        ";
	for (size_t j = 0; j < ao2._size; j++)
		os << " " << str11(ao2.uname(j));
	os << "\n\n";

	str8.left();

	double s02 = sigma0 * sigma0;
	int k = ao1._index;
	for (size_t i = 0; i < ao1._size; i++, k++)
		{
		os << str8(ao1.uname(i));
		int l = ao2._index;
		for (size_t j = 0; j < ao2._size; j++, l++)
			{
			double sij2 = qxx(k,l) * s02;
			os << " " << sci(sij2);
			}
		os << endl;
		}
	os << endl;
}

} // namespace DEP
