/* $Id: camera.cpp,v 1.18 2008/01/08 16:38:03 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "params.h"
#include "camera.h"
#include "imgpnt.h"
#include "apset.h"
#include "lobseqn.h"
#include "lconeqn.h"

namespace DEP
{

// constructor

Camera::Camera() :
	_id(),
	_comment(),
	_size_x(1.),
	_size_y(1.),
	_inuse(false),
	_ap_sets(),
	_princ_dist(1.),
	_princ_pnt(0., 0.)
{
	// empty
}


// get camera type

string Camera::type_str() const
{
	ostringstream oss;
	oss << "frame";
	return oss.str();
}


// get keyword name

string Camera::key_str() const
{
	ostringstream oss;
	oss << "camera";
	return oss.str();
}


// get coordinate system type

string Camera::csys_str() const
{
	ostringstream oss;
	oss << "image";
	return oss.str();
}


// read camera parameters

bool Camera::read(istream& is)
{
	// identifier
	is >> _id;
	if (!is) return false;

	// comment
	is >> _comment;
	if (!is) return false;

	// principal distance and principal point in mm
	is >> _princ_dist >> _princ_pnt;
	if (!is) return false;

	// convert into m
	_princ_dist *= 1e-3;
	_princ_pnt *= 1e-3;

	// image format in mm
	is >> _size_x >> _size_y;
	if (!is) return false;

	// convert into m
	_size_x *= 1e-3;
	_size_y *= 1e-3;

	return true;
}


// write camera parameters

void Camera::write(ostream& os) const
{
	Form fix6(6); fix6.fixed().width(12);

	os << "frame   " << _id << "   " << _comment << endl;
	os << fix6(_princ_dist * 1e3) << " "
		<< fix6(_princ_pnt.x * 1e3) << " " << fix6(_princ_pnt.y * 1e3) << endl;
	os << fix6(_size_x * 1e3) << " " << fix6(_size_y * 1e3) << endl;
}


// print camera parameters

void Camera::print(ostream& os) const
{
	Form fix4(4); fix4.fixed().width(10);

	os << "Camera ID          :  " << _id << endl;
	os << "Camera type        :  " << type_str() << endl;
	os << "Comment            :  " << _comment << endl;
	os << "Coord. system type :  " << csys_str() << endl;
	os << "Principal distance :  c  = " << fix4(_princ_dist * 1e3) << " mm\n";
	os << "Principal point    :  xp = " << fix4(_princ_pnt.x * 1e3) << " mm"
		"      yp = " << fix4(_princ_pnt.y * 1e3) << " mm\n";
	os << "Image format       :  sx = " << fix4(_size_x * 1e3) << " mm"
		"      sy = " << fix4(_size_y * 1e3) << " mm\n";
	os << endl;
}


// read sensor coords.

bool Camera::read_sensor_coords(istream& is,
vector<ImagePoint*>& impts, bool sdev_flg, const DPoint2& sdev) const
{
	// read image scale
	double sc = 1.;
	is >> sc;
	if (!is) return false;

	string id;

	while (is >> id)
		{
		if (id.compare("stop-dep") == 0 || id.compare("-1") == 0)
			break; // stop reading

		ImagePoint *ip = new ImagePoint;
		assert(ip != 0);
		ip->set_id(id);

		// read image coords.
		is >> ip->crds;
		if (!is) return false; // read error
		
		// std. devs.
		if (sdev_flg)
			ip->sdev = sdev;
		else
			{
			is >> ip->sdev;
			if (!is) return false;
			}
		
		// scale coords.
		ip->crds *= sc;
		ip->sdev *= sc;

		// append or replace point
		add_image_point(impts, ip);
		}

	return true;
}


// write sensor coords.

void Camera::write_sensor_coords(ostream& os,
const vector<ImagePoint*>& impts) const
{
	// scale/unit should be selectable by LS-params./options!
	const double sc = 1e-6;

	Form str8; str8.width(8);
	Form fix8(4); fix8.fixed().width(8);
	Form fix10(4); fix10.fixed().width(10);

	os << sc << endl;

	vector<ImagePoint*>::const_iterator ip;
	for (ip = impts.begin(); ip != impts.end(); ++ip)
		{
		if (*ip == 0) continue;
		os << str8((*ip)->get_id())
			<< " " << fix10((*ip)->crds.x * sc)
			<< " " << fix10((*ip)->crds.y * sc)
			<< " " << fix8((*ip)->sdev.x * sc)
				<< " " << fix8((*ip)->sdev.y * sc) << endl;
		}
	os << endl;
}


// print sensor coords.

void Camera::print_sensor_coords(ostream& os,
const vector<ImagePoint*>& impts, bool ic_out) const
{
	Form str8; str8.width(8);
	Form fix8(4); fix8.fixed().width(8);
	Form fix10(4); fix10.fixed().width(10);

	os << "Number of points        :  "
		<< DEP::count_image_points(impts) << endl;
	//os << "Coord. system type      :  " << csys_str() << endl;
	os << endl;
	
	if (ic_out == false) return; // escape
	
	os << "      ID     x/mm       y/mm     sx/mm    sy/mm\n\n";

	vector<ImagePoint*>::const_iterator ip;
	for (ip = impts.begin(); ip != impts.end(); ++ip)
		{
		if (*ip == 0) continue;
		os << str8((*ip)->get_id())
			<< " " << fix10((*ip)->crds.x * 1e3)
			<< " " << fix10((*ip)->crds.y * 1e3)
			<< " " << fix8((*ip)->sdev.x * 1e3)
				<< " " << fix8((*ip)->sdev.y * 1e3) << endl;
		}
	os << endl;
}


// image-to-photo-coords. transformation, AP correction included
// ATTENTION: the AP corrections are functions of the photo coords.!!!

void Camera::image_to_photo(const DPoint2& img_c, DPoint3& pho_c, bool corr_flg) const
{
	//cout << "Camera::image_to_photo" << endl;

	pho_c.x = img_c.x - _princ_pnt.x;
	pho_c.y = img_c.y - _princ_pnt.y;
	pho_c.z = -_princ_dist;
	//cout << pho_c << endl;

	// approximation only!!!
	if (corr_flg == true)
		{
		DPoint2 corr = ap_set_corr(pho_c);
		//cout << " AP corr:" << corr << endl;
		pho_c.x -= corr.x;
		pho_c.y -= corr.y;
		}

	// const double eps = 1e-8;
	// const int iter = 0;
	// const int max_iter = 10;
// 	do
// 		{
//		iter++;
// 		DPoint2 corr = ap_set_corr(pho_c0);
// 		pho_c.x = img_c.x - _princ_pnt.x - corr.x;
// 		pho_c.y = img_c.y - _princ_pnt.y - corr.y;
// 		cout << iter << ": " << pho_c << endl;
// 		pho_c0 = pho_c;
// 		}
// 		while ((pho_c.x - pho_c0.x) < eps && (pho_c.y - pho_c0.y) < eps && iter < max_iter)

// 	pho_c.x = img_c.x - _princ_pnt.x - corr.x;
// 	pho_c.y = img_c.y - _princ_pnt.y - corr.y;
// 	cout << pho_c << endl;
// 
// 		corr = ap_set_corr(pho_c); 
// 	pho_c.x = img_c.x - _princ_pnt.x - corr.x;
// 	pho_c.y = img_c.y - _princ_pnt.y - corr.y;
// 	cout << pho_c << endl;
// 
// 		corr = ap_set_corr(pho_c); 
// 	pho_c.x = img_c.x - _princ_pnt.x - corr.x;
// 	pho_c.y = img_c.y - _princ_pnt.y - corr.y;
// 	cout << pho_c << endl;
// 
// 		corr = ap_set_corr(pho_c); 
// 	pho_c.x = img_c.x - _princ_pnt.x - corr.x;
// 	pho_c.y = img_c.y - _princ_pnt.y - corr.y;
// 	cout << pho_c << endl << endl;
}


// photo-to-image-coords. transformation, AP correction included

void Camera::photo_to_image(const DPoint3& pho_c, DPoint2& img_c, bool corr_flg) const
{
	//cout << "Camera::photo_to_image" << endl;

	img_c.x = pho_c.x + _princ_pnt.x;
	img_c.y = pho_c.y + _princ_pnt.y;

	if (corr_flg == true)
		img_c += ap_set_corr(pho_c);
}


// get AP set ID
/*
string Camera::ap_set_id() const
{
	return (_ap_set == 0) ? (0) : (_ap_set->id());
}
*/

// AP set correction

DPoint2 Camera::ap_set_corr(const DPoint3& pho_c) const
{
	DPoint2 corr;
	
	vector<APSet*>::const_iterator aps;
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
	{
		if (*aps == 0) continue;
		corr += (*aps)->calc_corr(pho_c);
	}
	
	return corr;
}


// add subset of derivatives to linear observation equations

void Camera::add(LinObsEqn& eqn) const
{
	vector<APSet*>::const_iterator aps;
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
	{
		if (*aps == 0) continue;
		eqn.add_a(&(*aps)->_derivs);
	}
}


// add subset of derivatives to linear condition equations

void Camera::add_a(LinConditionEqn& eqn) const
{
	vector<APSet*>::const_iterator aps;
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
	{
		if (*aps == 0) continue;
		eqn.add_a(&(*aps)->_derivs);
	}
}


// get partial derivatives of IO parameters

void Camera::update_derivs(const DPoint3& pho_c)
{
	/*
	if (_ap_set != 0)
		{
		// get photo coords. from image coords.
		DPoint3 pho_c;
		image_to_photo(img_c, pho_c);
	
		_ap_set->get_derivs(pho_c, derivs);
		}
	*/
	
	vector<APSet*>::const_iterator aps;
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
	{
		if (*aps == 0) continue;
		(*aps)->update_derivs(pho_c);
	}
}


// get partial derivatives of IO parameters

void Camera::get_derivs(const DPoint3& pho_c, vector<DerivSub>& derivs) const
{
	assert(_ap_sets.size() == derivs.size());
	
	vector<DerivSub>::iterator drv = derivs.begin();
	
	vector<APSet*>::const_iterator aps;
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps, ++drv)
	{
		if (*aps == 0)
		{
			cout << "ERROR: Sensor::get_derivs: null pointer!" << endl;
			assert(false);
		}
		(*aps)->get_derivs(pho_c, *drv);
		drv->index((*aps)->get_index());
	}
}


// print influence of APs

void Camera::print_aps_influence(ostream& os, const Params& pr) const
{
	os << "Camera ID          :  " << _id << endl;
	os << "Camera type        :  " << type_str() << endl;
	os << "Comment            :  " << _comment << endl;
	os << endl;
	
	if (_ap_sets.empty())
		os << "  >>> no AP sets with this camera exist!" << endl;
	else
	{
		vector<APSet*>::const_iterator aps;
		for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
		{
			if (*aps == 0) continue;
			(*aps)->print_aps_influence(os, pr, *this); 
		}
	}
	os << endl;
}


// update IO parameter

bool Camera::update_io()
{
	bool rtn = false;
	
	vector<APSet*>::const_iterator aps;
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
	{
		if (*aps == 0) continue;
		
		if (APSetIO* p = dynamic_cast<APSetIO*>(*aps))
		{
			p->update(_princ_pnt.x, _princ_pnt.y, _princ_dist);
			rtn = true;
		}
	}
	
	return rtn;
}


// non-member functions ...

} // namespace DEP
