/* $Id: extori.h,v 1.17 2007/07/23 08:42:54 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_EXTORI_INCLUDED
#define DEP_EXTORI_INCLUDED

#include "station.h"
#include "drvrotmat.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{
class DerivSub;

/** Exterior orientation class.
 *  Base class for classes using the photogrammetric model for frame camera.
 *
 *  The spatial position and attitude of the camera
 *  station is defined by the Exterior Orientation (EO).
 *  The spatial position, also called perspective projection centre, is
 *  defined by its coordinates Xo, Yo and Zo. The attitude is defined by
 *  the rotation matrix. In this base class no assumtion is made on
 *  the parametrisation of the rotation matrix.
 *  Different parametrisations exist. Usual parametrisations are:
 *  three rotation angles (euler angles), unit quaternion, Rodriguez matrix, 
 *  or direction cosines (the rotation matrix itself).
 *  This spezialisation is done by the drived classes,
 *  e.g. EOAnglesOPKRot, EOQuaternion.
 *
 *  @author Dirk Stallmann
 */
class ExtOrientation : public Station
{
// 	friend double intersection(const ExtOrientation& eo1,
// 		const DPoint3& pho_c1, const ExtOrientation& eo2,
// 		const DPoint3& pho_c2, DPoint3& obj_c);

public:
	/** constructor
	 *  @param sz number of rotation parameters (3 <= sz <= 9)
	 */
	ExtOrientation(size_t sz=3);

	/** destructor
	 */
	~ExtOrientation();
	
	/**@name read/write/print EO parameters */
	//@{
		/// read EO parameters
		virtual bool read(istream& is);

		/// write EO parameters
		virtual void write(ostream& os) const;

		/// print EO parameters
		virtual void print(ostream& os) const;
	//@}

	/// set and get perspective centre coords.
	//@{
		/// set perspective centre coords.
		void set_per_centre(const DPoint3& pcc)
		{
			_per_centre = pcc;
		}
		
		/// perspective centre coords.
		void get_per_centre(DPoint3& pcc) const
		{
			pcc = _per_centre;
		}
		
		/// set std. devs.
		void set_sdev_per_centre(const DPoint3& sdev)
		{
			_sdev_pcc = sdev;
		}
	//@}

	/// set and get rotation parameters
	//@{
		/// get number of rotation parameters
		size_t num_rot_params() const { return _a.size(); }
		
		/// set rotation parameters
		void set_rot_params(const vector<double>& a)
		{
			_a = a;
		}

		/// get rotation parameters
		void get_rot_params(vector<double>& a) const
		{
			a = _a;
		}
		
		/// set std. devs.
		void set_sdev_rot_params(const vector<double>& sdev)
		{
			_sdev_a = sdev;
		}
	//@}
	
	/// set and get exposure time
	//@{
		/// set exposure time
		void set_time(double t) { _time = t; }

		/// get exposure time
		double get_time() const { return _time; }
	//@}
	
	/// pure virtual functions
	//@{
		/// calculate rotation parameters from rotation matrix
		virtual void calc_rot_params(vector<double>& a, const DRotMatrix3& R) const = 0;
		
		/// calculate rotation matrix
		virtual void calc_rot_matrix(DRotMatrix3& R, const vector<double>& a) const = 0;
		
		/** calculate derivatives of rotation parameters
		 *   @param deriv (OUT) partial derivatives
		 *   @param h (IN) rotation matrix between object and photo coordinate system
		 *   @param u (IN) auxillary variable
		 *   @param v (IN) auxillary variable
		 *   @param w (IN) auxillary variable
		 *   @param d (IN) object coords. reduced to projection centre
		 */
		virtual void calc_derivs_rot(DerivSub& deriv,
			double h, double u, double v, double w,
			const DPoint3& d, const vector<double>& a) const = 0;
		
		/** calculate derivatives of rotation parameters
		 *   @param deriv (OUT) partial derivatives
		 *   @param h (IN) rotation matrix between object and photo coordinate system
		 *   @param u (IN) auxillary variable
		 *   @param v (IN) auxillary variable
		 *   @param w (IN) auxillary variable
		 *   @param d (IN) object coords. reduced to projection centre
		 *   @param M (IN) misalignment rotation matrix
		 */
		virtual void calc_derivs_rot(DerivSub& deriv,
			double h, double u, double v, double w,
			const DPoint3& d, const DRotMatrix3& M, const vector<double>& a) const = 0;
	//@}
	
	/**@name overridden methods of base class AdjObject
	 */
	//@{
		/// get adjustment object class name (long description)
		string class_str() const;
		
		/// set unknown parameter information (upari)
		bool set_upari();
		
		/// print changes of unknowns
		virtual void print_changes(ostream& os, const Vector<double>& dx) const;

		/// print unknown parameters
		virtual void print_unknowns(ostream& os) const;
		
		/// check on convergence
		int check_convergence(const Vector<double>& dx) const;
	//@}
	
	/** Sum of absolute changes
	 *
	 *  Overwritten base class function. Simply do nothing. 
	 *  The summation is done separatly for the perspective centre coords.
	 *  and the rotation parameters.
	 */
	//@{
		/// sum of absolute changes.
		void sum_changes(const Vector<double>& dx,
			double& sum, size_t& num) const {}

		/// sum of absolute changes of perspective centre coords.
		void sum_changes_pcc(const Vector<double>& dx,
			double& sum, size_t& num) const;

		/// sum of absolute changes of rotation parameters
		void sum_changes_rot(const Vector<double>& dx,
			double& sum, size_t& num) const;
	//@}
	
	// spatial intersection
	bool intersection(const ExtOrientation& eo2,
	const DPoint3& pho_c1, const DPoint3& pho_c2,
	DPoint3& obj_c, double& e) const;
	
protected:
	/// exposure/reference time
	double	_time;
	
	/// perspective centre coordinates
	DPoint3	_per_centre;
	
	/// std. dev. of perspective centre coordinates
	DPoint3	_sdev_pcc;
	
	/// rotation parameters
	vector<double>	_a;

	/// std. dev. of rotation parameters
	vector<double>	_sdev_a;
};


// non-member functions ...

/** Project object point onto photo plane.
 *  Basic collinearity equation to get photo from object coordinates.
 *
 *   @param pc (IN) perspective centre coords.
 *   @param rm (IN) rotation matrix between object and photo coordinate system
 *   @param obj_c (IN) object point coords.
 *   @param pho_c photo coords.: pho_c.x, pho_c.y (OUT) x,y photo coords.,
 *        pho_c.z (IN) c principal distance
 *   @return true = no problem, false = error encountered
 */
bool projection(const DPoint3& pc, const DRotMatrix3& rm,
	const DPoint3& obj_c, DPoint3& pho_c);

/** Spacial intersection.
 *  Intersection by using the minimal distance e between two rays.
 *  Method published by Albertz/Kreiling.
 *
 *   @param pc1 (IN) perspective centre coords. of image 1
 *   @param rm1 (IN) rotation matrix of image 1
 *   @param pho_c1 (IN) photo coords. in image 1
 *   @param rm2 (IN) rotation matrix of image 2
 *   @param pc2 (IN) perspective centre coords. of image 2
 *   @param pho_c2 (IN) photo coords. in image 1
 *   @param obj_c (OUT) object point coords.
 *   @param e (OUT) minimal distance e
 *   @return true = no problem, false = error encountered
 */
bool intersection(
	const DPoint3& pc1, const DRotMatrix3& rm1, const DPoint3& pho_c1,
	const DPoint3& pc2, const DRotMatrix3& rm2, const DPoint3& pho_c2,
	DPoint3& obj_c, double& e);

/** ray-z-plane intersection.
 *  Project image point onto Z-plane.
 *
 *   @param pc (IN) perspective centre coords. of image
 *   @param rm (IN) rotation matrix of image
 *   @param obj_c (IN) object point coords.
 *   @param pho_c (OUT) photo coords. in image
 */
 void intersection(const DPoint3& pc, const DRotMatrix3& rm,
	DPoint3& obj_c, const DPoint3& pho_c);

} // namespace DEP

#endif
