/* $Id: fiodep.cpp,v 1.24 2007/12/10 06:56:44 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <fstream>
#include "depvec.h"
#include "fiodep.h"
#include "lspars.h"
#include "scanner.h"
#include "imgpnt.h"
#include "objpnt.h"
#include "ccdcamtr.h"
#include "pbcam.h"
#include "adscam.h"
#include "eopokrot.h"
#include "eoopkrot.h"
#include "eoopkfix.h"
#include "eoaustralis.h"
#include "eoquaternion.h"
#include "caminscfg.h"
#include "dlt.h"
//#include "eokratky.h"
#include "pmflin.h"
//#include "eodirobs.h"
#include "apset.h"
#include "poscor.h"
#include "attcor.h"
#include "adsorient.h"
#include "imuopkfix.h"
#include "imuquat.h"

namespace DEP
{

// enum, keywords

FileIO::map_t FileIO::_map[] =
{
	{ UNKNOWN,			"unknown" },
	{ PARAMS,			"params" },
	{ LS_PARAMS,		"ls-params" },
	{ NETWORK_DESIGN_PARAMS,		"network-design-params" },
	{ CAMERA,			"camera" },
	{ CAMERA_FRAME,		"camera-frame" },
	{ CAMERA_CCD,		"camera-ccd" },
	{ CAMERA_CCD_TR,	"camera-ccd-tr" },
	{ CAMERA_PUSHBROOM,	"camera-pushbroom" },
	{ CAMERA_PUSHBROOM_CALIB,	"camera-pushbroom-calib" },
	{ CAMERA_ADS,		"camera-ads" },
	{ IMAGE,			"image" },
	{ IMAGE_FRAME,		"image-frame" },
	{ IMAGE_FRAME_EXT,		"image-frame-ext" },
	{ IMAGE_SCANNER,	"image-scanner" },
	{ IMAGE_DLT,		"image-dlt" },
	{ IMAGE_AFFINE_TRANSF,		"image-affine-transf" },
	{ STATION,			"station" },
	{ EXT_ORI,			"ext-ori" },
	{ EXT_ORI_OPK_ROT,	"ext-ori-opk-rot" },
	{ EXT_ORI_POK_ROT,	"ext-ori-pok-rot" },
	{ EXT_ORI_OPK_FIX,	"ext-ori-opk-fix" },
	{ EXT_ORI_AUSTRALIS,	"ext-ori-australis" },
	{ EXT_ORI_QUATERNION,	"ext-ori-quaternion" },
	{ DLT_PARAMS,			"dlt-params" },
	{ AFFINE_TRANSF_PARAMS,	"affine-transf-params" },
	{ IMAGE_COORDS,		"image-crds" },
	{ OBJECT_COORDS,	"obj-crds" },
	{ CONTROL_SUPP,		"ctrl-supp" },
	{ APSET,			"ap-set" },
	{ APSET_GAP,		"gap" },
	{ APSET_AUSTRALIS,	"australis" },
	{ APSET_NSE,		"nse" },
	{ APSET_P12,		"p12" },
	{ APSET_P44,		"p44" },
	{ APSET_INNER_OR,	"inner-or" },
	{ APSET_RADIAL,		"radial-dist" },
	{ APSET_DECENTERING,	"decentering-dist" },
	{ APSET_INPLANE,	"in-plane-dist" },
	{ APSET_OUTOFPLANE,	"out-of-plane-dist" },
	{ APSET_BROWN_EQN22,	"brown-eqn-22" },
	{ APSET_BROWN_EQN23,	"brown-eqn-23" },
	{ POS_COR,			"pos-cor" },
	{ ATT_COR,			"att-cor" },
	{ ORIENTATION_DATA_PARAMS,		"orientation-data-params" },
	{ ORIENTATION_DATA,				"orientation-data" },
	{ ORIENTATION_DATA_ODF,			"orientation-data-odf" },
	{ CAM_INS_CONFIG,	"cam-ins-config" },
	{ IMU_MISALIGNMENT,	"imu-misal" },
	{ PATB_PARAMS,		"patb-params" },
	{ PATB_PHOTOGRAPHS,	"patb-photos" },
	{ PATB_CONTROL_POINTS,	"patb-control" },
	{ PATB_ORIENT_PARAMS,	"patb-orient" },
	{ OC_TRANSFORM,		"oc-transform" },
	{ STOP_DEP,			"stop-dep" },
	{ END_OF_MAP,		0 }
};


// get code

FileIO::ecode FileIO::code(const char* keyword)
{
	for (int i = 0; _map[i].keyword != 0; i++)
		{
		if (strcmp(keyword, _map[i].keyword) == 0)
			return (ecode)_map[i].code;
		}

	return (ecode)_map[0].code;
}


// determine file type

bool FileIO::file_type(const char* fname, ecode& type) const
{
	// open file
	ifstream is(fname, ios_base::binary);
	if (!is) return false;
	
	// read magic number
	char magco[3];
	is.get(magco[0]);
	is.get(magco[1]);
	is.get(magco[2]);
	if (!is) return false;

	// check magic number
	if (magco[0] == 'D' && magco[1] == 'E' && magco[2] == 'P')
		{
		// DEP format, determine sub-type
		string keyword;
		is >> keyword;
		if (!is) return false;
		type = code(keyword.c_str());
		}
	else if (magco[0] == 'O' && magco[1] == 'D' && magco[2] == 'F')
		type = ORIENTATION_DATA_ODF; // ODF format
	else
		type = UNKNOWN;
	
	return true;
}


// read cameras

bool FileIO::read_cameras(istream& is, vector<Camera*>& cams) const
{
	string str;

	while (is >> str)
		{
		ecode type = code(str.c_str());
		if (type == STOP_DEP) break;

		// create new camera
		Camera* cam = 0;
		switch (type)
			{
			case CAMERA_FRAME: cam = new Camera; break;
			case CAMERA_CCD: cam = new CCDCamera; break;
			case CAMERA_CCD_TR: cam = new CCDCameraTR; break;
			case CAMERA_PUSHBROOM: cam = new PBCamera; break;
			case CAMERA_PUSHBROOM_CALIB: cam = new PBCameraCalib; break;

			case UNKNOWN:
			default:
				cout << "ERROR: Unknown camera type: " << str << endl;
				return false;
			}
		assert(cam != 0);

		// read camera parameter
		if (cam->read(is) != true)
			{
			cout << "ERROR: Reading camera " << cam->get_id() << " failed.\n";
			delete cam;
			return false;
			}
		
		// append or replace camera
		if (append(cams, cam) == false)
			cout << "ATTENTION: Camera " << cam->get_id() << " was replaced.\n";
		}

	return true;
}


// read images

bool FileIO::read_images(istream& is, const vector<Camera*>& cams,
vector<Image*>& images)
{
	string str;

	while (is >> str)
		{
		ecode type = code(str.c_str());
		if (type == STOP_DEP) break;

		// create new image
		Image* image = 0;
		switch (type)
			{
			case IMAGE_FRAME: image = new Image; break;
			case IMAGE_FRAME_EXT: image = new ExtFrameImage; break;
			case IMAGE_SCANNER: image = new ScannerImage; break;
			case IMAGE_DLT: image = new DLTImage; break;
			case IMAGE_AFFINE_TRANSF: image = new AffineTransfImage; break;

			case UNKNOWN:
			default:
				cout << "ERROR: Unknown image type: " << str << endl;
				return false;
			}
		assert(image != 0);

		// read image information
		if (image->read(is, cams) != true)
			{
			cout << "ERROR: Reading image " << image->get_id() << " failed.\n";
			delete image;
			return false;
			}

		// append or replace
		if (append(images, image) == false)
			cout << "ATTENTION: Image " << image->get_id()
				<< " was replaced.\n";
		}

	return true;
}


// read stations

bool FileIO::read_stations(istream& is, vector<Station*>& stats)
{
	string str;

	if (!(is >> str))
		{
		cout << "ERROR: Reading keyword failed.\n";
		return false;
		}

	if (str.compare("indiv-type") == 0)
		{
		if (read_stations_indiv(is, stats) != true)
			{
			cout << "ERROR: Reading station with individual type failed.\n";
			return false;
			}
		}
	else if (str.compare("common-type") == 0)
		{
		if (read_stations_common(is, stats) != true)
			{
			cout << "ERROR: Reading stations with common type failed.\n";
			return false;
			}
		}
	else
		{
		cout << "ERROR: Unknown keyword: " << str << endl;
		return false;
		}

	return true;
}


// read stations have individual type and std. devs.

bool FileIO::read_stations_indiv(istream& is, vector<Station*>& stats)
{
	string str;

	while (is >> str)
		{
		ecode type = code(str.c_str());
		if (type == STOP_DEP) break;

		// create new image geometry
		Station* eo = 0;
		switch (type)
			{
			case EXT_ORI_OPK_ROT: eo = new EOAnglesOPKRot; break;
			case EXT_ORI_POK_ROT: eo = new EOAnglesPOKRot; break;
			case EXT_ORI_OPK_FIX: eo = new EOAnglesOPKFix; break;
			case EXT_ORI_AUSTRALIS: eo = new EOAustralis; break;
			case EXT_ORI_QUATERNION: eo = new EOQuaternion; break;
			case DLT_PARAMS: eo = new DLTParameters; break;
			case AFFINE_TRANSF_PARAMS: eo = new AffineTransfParams; break;
			
			case UNKNOWN:
			default:
				cout << "ERROR: Unsupported station type: "
					<< str << endl;
				return false;
			}
		assert(eo != 0);

		// read station data
		if (eo->read(is) != true)
			{
			cout << "ERROR: Reading station failed.\n";
			delete eo;
			return false;
			}

		// append or replace
		if (append(stats, eo) == false)
			cout << "ATTENTION: Station " << eo->get_id() << " was replaced.\n";
		}

	return true;
}


// read EO have common type and std. devs.

bool FileIO::read_stations_common(istream& is, vector<Station*>& stats)
{
	string str;

	// read type of geometry
	is >> str;
	if (!is)
		{
		cout << "ERROR: Cannot read station type.\n";
		return false;
		}

	// type of station
	ecode type = code(str.c_str());

	// check and create EO
	ExtOrientation *eo = 0;
	switch (type)
		{
		// valid EOs
		case EXT_ORI_OPK_ROT: eo = new EOAnglesOPKRot; break;
		case EXT_ORI_POK_ROT: eo = new EOAnglesPOKRot; break;
		case EXT_ORI_OPK_FIX: eo = new EOAnglesOPKFix; break;
		case EXT_ORI_AUSTRALIS: eo = new EOAustralis; break;
		
		case UNKNOWN:
		default:
			cout << "ERROR: Unsupported station type: " << str << endl;
			return false;
		}
	assert(eo != 0);

	// unit name of rotation angles
	is >> str;
	if (!is)
		{
		cout << "ERROR: Cannot read unit of rotation angles.\n";
		return false;
		}

	UnitAngle u_angle;
	if (!u_angle.set(str.c_str()))
		{
		cout << "ERROR: Invalid unit of rotation angles: "
			<< str << endl;
		return false;
		}
	double rho = u_angle.rho();

	// time flag
	bool time_flg = true;
	is >> time_flg;
	if (!is)
		{
		cout << "ERROR: Cannot read time flag.\n";
		return false;
		}

	// rotation matrix flag:
	bool matrix_flg = false;
	is >> matrix_flg;
	if (!is)
		{
		cout << "ERROR: Cannot read rotation matrix flag.\n";
		return false;
		}

	// read common std. dev.
	DPoint3 sdev_pcc;
	is >> sdev_pcc;
	if (!is)
		{
		cout << "ERROR: Cannot read common std. devs. of per. centre coords.\n";
		return false;
		}
	
	// std. dev. of rotation parameters
	vector<double> sdev_a(eo->num_rot_params());
	
	for (size_t i = 0; i < sdev_a.size(); i++)
		{
		is >> sdev_a[i];
		if (!is)
			{
			cout << "ERROR: Cannot read common std. devs. of rotation parameters failed.\n";
			return false;
			}
			
		// only for real std. devs. conversion into radian
		if (AdjObject::lspars.is_rsdev(sdev_a[i])) sdev_a[i] /= rho;
		}

	string		id;			// station ID
	double		time = 0.;	// exposure time
	DPoint3		pcc;		// perspective centre coords.
	vector<double>	a(eo->num_rot_params());
	DRotMatrix3	rm;			// rotation matrix

	while (is >> id)
		{
		if (id.compare("stop-dep") == 0 || id.compare("-1") == 0) break;

		// create EO
		switch (type)
			{
			case EXT_ORI_OPK_ROT: eo = new EOAnglesOPKRot; break;
			case EXT_ORI_POK_ROT: eo = new EOAnglesPOKRot; break;
			case EXT_ORI_OPK_FIX: eo = new EOAnglesOPKFix; break;
			case EXT_ORI_AUSTRALIS: eo = new EOAustralis; break;

			default:
				assert(false);
			}
		assert(eo != 0);
		eo->set_id(id);
		
		if (time_flg)
			{
			if (!(is >> time))
				{
				cout << "ERROR: Cannot read exposure time.\n";
				return false;
				}
			}
		eo->set_time(time);
		
		if (!(is >> pcc))
			{
			cout << "ERROR: Cannot read per. centre coords.\n";
			return false;
			}
		eo->set_per_centre(pcc);
		
		eo->set_sdev_per_centre(sdev_pcc);
		
		if (matrix_flg == false)
			{
			for (size_t i = 0; i < a.size(); i++)
				{
				is >> a[i];
				if (!is)
					{
					cout << "ERROR: Cannot read rotation parameters.\n";
					return false;
					}
				a[i] /= rho; // conversion into radian
				}
			}
		else
			{
			if (!(is >> rm))
				{
				cout << "ERROR: Cannot read rotation matrix.\n";
				return false;
				}
			//rm.transpose();
			eo->calc_rot_params(a, rm);
			}
		eo->set_rot_params(a);
		eo->set_sdev_rot_params(sdev_a);
		eo->set_approx(true);

		// append or replace
		if (append(stats, dynamic_cast<Station*>(eo)) == false)
			cout << "ATTENTION: Station " << id << " was replaced.\n";
		}

	return true;
}


// read image coordinates

bool FileIO::read_image_coords(istream& is, vector<Image*>& images)
{
	DPoint2 sdev;
	string id;
	string str;
	
	bool sdev_flg = false;

	is >> str;
	if (!is) return false;

	if (str.compare("indiv-sdev") == 0)
		sdev_flg = false;
	else if (str.compare("common-sdev") == 0)
		{
		sdev_flg = true;
		is >> sdev;
		}
	else
		{
		cout << "ERROR: Unknown file type: " << str << endl;
		return false;
		}
	
	while (is >> id)
		{
		if (id.compare("stop-dep") == 0 || id.compare("-1") == 0) break;

		// find image
		Image* image = find(images, id);
		if (image == 0)
			{
			cout << "ERROR: Image " << id << " not found.\n";
			return false;
			}

		// append or replace image coords.
		if (image->read_image_coords(is, sdev_flg, sdev) != true)
			{
			cout << "ERROR: Image " << id << " read image coords. failed.\n";
			return false;
			}
		}

	return true;
}


// read object point coords.

bool FileIO::read_object_coords(istream& is,
vector<ObjectPoint*>& obpts)
{
	DPoint3 crds, sdev;
	string id;
	string str;
	
	bool com_sdev = false;

	is >> str;
	if (!is) return false;

	if (str.compare("indiv-sdev") == 0)
		com_sdev = false;
	else if (str.compare("common-sdev") == 0)
		com_sdev = true;
	else
		{
		cout << "ERROR: Unknown file type: " << str << endl;
		return false;
		}

	// read common std. dev.
	if (com_sdev)
		{
		is >> sdev;
		if (!is)
			{
			cout << "ERROR: Reading common std. dev. failed.\n";
			return false;
			}
		}

	while (is >> id)
		{
		if (id.compare("stop-dep") == 0 || id.compare("-1") == 0)
			break; // stop reading

		// read coords.
		is >> crds;
		if (!is)
			{
			cout << "ERROR: Reading coords. failed.\n";
			return false;
			}

		// read individual std. dev.
		if (!com_sdev)
			{
			is >> sdev;
			if (!is)
				{
				cout << "ERROR: Reading std. dev. failed.\n";
				return false;
				}
			}
			
		ObjectPoint* op = find(obpts, id);
		if (op != 0)
			cout << "ATTENTION: Object point " << id << " replaced.\n";
		else
			{
			// create new object point
			op = new ObjectPoint;
			assert(op != 0);

			op->set_id(id);
			obpts.push_back(op);
			}

		// store coords. and std. dev.
		op->crds = crds;
		op->sdev = sdev;
		op->set_approx(true);

		// set type from std. dev.
		// Attention: type and std. devs. must be correspond!!!
		op->set_type();
		}

	return true;
}


// read control support definition

bool FileIO::read_ctrl_supp(istream& is, vector<ObjectPoint*>& obpts)
{
	for (;;)
		{
		ObjectPointCS cs;

		// read CS
		if (cs.read_ctrl_supp(is) != true)
			{
			cout << "ERROR: Set std. dev. from CS failed.\n";
			return false;
			}

		if (!is ||
			cs.id.compare("stop-dep") == 0 || cs.id.compare("-1") == 0)
			break; // stop reading

		// find object point
		ObjectPoint* op = find(obpts, cs.id);
		if (op == 0)
			{
			cout << "ERROR: Object point ID " << cs.id << " not exist.\n";
			return false;
			}

		// set std. dev. from CS
		if ((op->set_sdev(cs)) != true)
			{
			cout << "ERROR: Set std. dev. from CS failed: " << cs.id << endl;
			return false;
			}
		}

	return true;
}


// read AP sets

bool FileIO::read_ap_sets(istream& is, vector<APSet*>& aps)
{
	string str;

	while (is >> str)
		{
		ecode type = code(str.c_str());
		if (type == STOP_DEP) break;

		// create new AP set
		APSet* ap = 0;
		switch (type)
			{
			case APSET_GAP:			ap = new APSetGAP; break;
			case APSET_AUSTRALIS:	ap = new APSetAustralis; break;
			case APSET_NSE:			ap = new APSetNSE; break;
			case APSET_P12:			ap = new APSetP12; break;
			case APSET_P44:			ap = new APSetP44; break;
			case APSET_INNER_OR:	ap = new APSetIO; break;
			case APSET_RADIAL:		ap = new APSetRadial; break;
			case APSET_DECENTERING: ap = new APSetDecentering; break;
			case APSET_INPLANE:		ap = new APSetInPlane; break;
			case APSET_OUTOFPLANE:	ap = new APSetOutOfPlane; break;
			case APSET_BROWN_EQN22:	ap = new APSetBrownEqn22; break;
			case APSET_BROWN_EQN23:	ap = new APSetBrownEqn23; break;

			case UNKNOWN:
			default:
				cout << "ERROR: Read AP set: Unknown type: " << str << endl;	
				return false;
			}
		assert(ap != 0);

		// read ap set
		//if (ap->read(is, lspars) != OK)
		if (ap->read(is) != true)
			{
			cout << "ERROR: Read AP set " << ap->get_id() << " failed.\n";
			delete ap;
			return false;
			}

		// append or replace
		if (append(aps, ap) == false)
			cout << "ATTENTION: AP set " << ap->get_id() << " was replaced.\n";
		}

	return true;
}


// read position corrections

bool FileIO::read_pos_correction(istream& is,
vector<PosCorrection*>& pcors)
{
	string id;

	while (is >> id)
		{
		if (id.compare("stop-dep") == 0) break;

		// create new object
		PosCorrection* pcor = new PosCorrection;
		assert(pcor != 0);

		// set ID
		pcor->set_id(id);

		// read data
		if (pcor->read(is) != true)
			{
			cout << "ERROR: Read position correction " << id << " failed.\n";
			delete pcor;
			return false;
			}

		// append or replace
		if (append(pcors, pcor) == false)
			cout << "ATTENTION: Position correction " << id
				<< " was replaced.\n";
		}

	return true;
}


// read attitude corrections

bool FileIO::read_att_correction(istream& is,
vector<AttCorrection*>& acors)
{
	string id;

	while (is >> id)
		{
		if (id.compare("stop-dep") == 0) break;

		// create new object
		AttCorrection* acor = new AttCorrection;
		assert(acor != 0);

		// set ID
		acor->set_id(id);

		// read data
		if (acor->read(is) != true)
			{
			cout << "ERROR: Read attitude correction " << id << " failed.\n";
			delete acor;
			return false;
			}

		// append or replace
		if (append(acors, acor) == false)
			cout << "ATTENTION: Attitude correction " << id
				<< " was replaced.\n";
		}

	return true;
}


// read IMU misaligment data

bool FileIO::read_imu_misalignment(istream& is,
vector<IMUMisalignment*>& v)
{
	string str;

	while (is >> str)
		{
		ecode type = code(str.c_str());
		if (type == STOP_DEP) break;

		ERotMatrix3 etype;
		ERotMatrix3::ecode rtype = etype.code(str.c_str());
		
		// create new object
		IMUMisalignment* p = 0;
		switch (rtype)
			{
			case ERotMatrix3::ANGLES_OPK_FIX:
				p = new IMUMisalignmentOPKFix; break;
			case ERotMatrix3::UNIT_QUATERNION:
				p = new IMUMisalignmentQuaternion; break;
			
			case UNKNOWN:
			default:
				cout << "ERROR: Read IMU misaligment data: Unknown type: "
					<< str << endl;
				return false;
			}
		assert(p != 0);

		// read data
		if (p->read(is) != true)
			{
			cout << "ERROR: Read IMU misaligment data " << p->get_id()
				<< " failed.\n";
			delete p;
			return false;
			}

		// append or replace
		if (append(v, p) == false)
			cout << "ATTENTION: IMU misaligment data " << p->get_id()
				<< " was replaced.\n";
		}

	return true;
}


// read orientation data sets

bool FileIO::read_orientation_data(const char* fname, ecode type,
const char* id, vector<OrientationData*>& ods) const
{
	// create new object
	OrientationData* p = 0;
	switch (type)
		{
		case ORIENTATION_DATA:		p = new OrientationData; break;
		case ORIENTATION_DATA_ODF:	p = new ADSOrientation; break;
		
		case UNKNOWN:
		default:
			cout << "ERROR: Read orientation data: Unknown type: "
				<< keyword(type) << endl;
			return false;
		}
	assert(p != 0);

	// read data
	if (!p->read(fname))
		{
		cout << "ERROR: Read orientation data failed: " << fname << endl;
		delete p;
		return false;
		}
	
	if (dynamic_cast<ADSOrientation*>(p) != 0) p->set_id(id); // set ID

	// append or replace
	if (append(ods, p) == false)
		cout << "ATTENTION: Orientation data set " << p->get_id()
			<< " was replaced.\n";

	return true;
}


// read ADS camera

bool FileIO::read_ads_camera(const char* fname, ecode type,
const char* id, vector<Camera*>& cams) const
{
	// create new camera
	Camera* p = 0;
	switch (type)
		{
		case CAMERA_ADS:	p = new ADSCamera; break;

		case UNKNOWN:
		default:
			cout << "ERROR: Unknown camera type: " << keyword(type) << endl;
			return false;
		}
	assert(p != 0);

	// read camera data
	ifstream is(fname);
	if (!is)
		{
		cout << "ERROR: Cannot open camera file: " << fname << endl;
		delete p;
		return false;
		}
		
	if (p->read(is) != true)
		{
		cout << "ERROR: Reading camera failed: " << fname << endl;
		delete p;
		return false;
		}
	
	p->set_id(id); // set ID
		
	// append or replace camera
	if (append(cams, p) == false)
		cout << "ATTENTION: Camera " << p->get_id() << " was replaced.\n";

	return true;
}

} // namespace DEP
