/* $Id: image.cpp,v 1.22.2.1 2008/03/05 18:25:05 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "depvec.h"
#include "image.h"
#include "lspars.h"
#include "drvsub.h"
#include "lconeqn.h"
#include "imgpnt.h"
#include "objpnt.h"
#include "camera.h"
#include "extori.h"
#include "apset.h"
#include "poscor.h"
#include "attcor.h"
#include "iopntcon.h"

namespace DEP
{

// constructor

Image::Image() :
	_image_id(),
	_station_id(),
	_station(0),
	_camera_id(),
	_camera(0),
	_ap_sets(0),
	_inuse(false),
	_pos_cor(0),
	_att_cor(0)
{
	// empty
}


// read image configuration

bool Image::read(istream& is, const vector<Camera*>& cams)
{
	// read image, station and camera ID
	is >> _image_id >> _station_id >> _camera_id;
	if (!is) return false;

	// attach camera
	assert(_camera == 0); // assume camera is not defined!

	if (!attach(cams))
		{
		cout << "ERROR: Attach camera " << _camera_id
			<< " for image " << _image_id << " failed.\n";
		return false;
		}

	return true;
}


// print image configuration

void Image::print(ostream& os) const
{
	os << "Image ID             :  " << _image_id << endl;
	os << "Image type           :  " << type_str() << endl;
	os << "Station ID           :  " << _station_id << endl;
	os << "Camera ID            :  " << _camera_id << endl;
	/*
	os << "AP set IDs          :  ";
	if (_ap_sets.empty())
		os << "no sets defined" << endl;
	else
		{
		os <<
		vector<APSet*>::const_iterator aps;
		for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
			os << "  " << (*aps)->get_id() << endl;
	 	}
	*/
	os << endl;
}


// get image type

string Image::type_str() const
{
	ostringstream oss;
	oss << "frame";
	return oss.str();
}


// read image coords.

bool Image::read_image_coords(istream& is, bool sdev_flg, const DPoint2& sdev)
{
	assert(_camera != 0);
	
	return _camera->read_sensor_coords(is, _image_points, sdev_flg, sdev);
}


// write image coords.

void Image::write_image_coords(ostream& os) const
{
	assert(_camera != 0);

	os << _image_id << "   ";
	_camera->write_sensor_coords(os, _image_points);
}


// print image coords.

void Image::print_image_coords(ostream& os) const
{
	assert(_camera != 0);
	
	os << "Image ID :  " << _image_id << endl;
	_camera->print_sensor_coords(os, _image_points, true);
}


// add (append or replace) image point

bool Image::add_image_point(ImagePoint* ip)
{
	assert(ip != 0);
	bool rtn = false;

	int i = find_image_point(ip->_id);
	if (i < 0)
		{
		_image_points.push_back(ip);
		rtn = true;
		}
	else
		{
		ImagePoint* tmp = _image_points[i];
		_image_points[i] = ip;
		delete tmp;
		cout << "ATTENTION: Image " << _image_id
			<< " point " << ip->_id << " was replaced.\n";
		}

	return rtn;
}


// count number of image points

int Image::count_image_points() const
{
	int n = 0;

	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip != 0) n++;
		}

	return n;
}


// test image point

bool Image::image_point_exist(const string& id) const
{
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip != 0 && (*ip)->get_id() == id) return true;
		}

	return false;
}


// find image point by id, return index

int Image::find_image_point(const string& id) const
{
	int i = 0;

	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip != 0 && (*ip)->get_id() == id) return i;
		i++;
		}

	return -1;
}


// get image point by id, return pointer

ImagePoint* Image::get_image_point(const string& id) const
{
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		if (*ip != 0 && (*ip)->get_id() == id) return *ip;

	return 0;
}


// remove image point
// erase by set element to 0!!!

bool Image::remove_image_point(const string& id)
{
	vector<ImagePoint*>::iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip != 0 && (*ip)->get_id() == id)
			{
			delete *ip;
			*ip = 0;
			return true;
			}
		}

	return false;
}


// write adjusted image coords.
/*
void Image::write_adjust_image_coords(ostream& os) const
{
	assert(_camera != 0);

	Form str8; str8.width(8);
	Form fix3(3); fix3.fixed().width(10);

	const double sc = 1e6; // convert in mue-meter
	DPoint2 img_c0, img_c;

	os << _image_id << endl;

	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;

		// find corresponding object point
		ObjectPoint* op = find(obj_pnts, (*ip)->get_id());
		if (op == 0)
			{
			cout << "ATTENTION: Object point "
				<< (*ip)->get_id() << " not found" << endl;
			continue;
			}

		// collinear equation -> calculate image coords.
		object_to_image(op->crds, img_c0);

		// convert sensor into image coords.
		_camera->sensor_to_image((*ip)->crds, img_c);

		// residuals: Soll - Ist
		v = img_c0 - img_c;

		os << "  " << str8((*ip)->get_id()) << "  "
			<< fix3(v.x * sc) << " " << fix3(v.y * sc) << endl;
		}
	os << "stop-dep\n";
}
*/


// attach station

bool Image::attach(const vector<Station*>& stats)
{
	if (_station_id.empty() || _station_id.compare("none") == 0) return true; // ignored
	
	_station = find(stats, _station_id);
	if (_station == 0)
		{
		cout << "ERROR: Attach station " << _station_id
			<< " for image " << _image_id << " failed.\n";
		return false;
		}
		
	_station->inuse(true);
	
	return true;
}


// attach camera

bool Image::attach(const vector<Camera*>& cams)
{
	_camera = find(cams, _camera_id);
	if (_camera == 0)
		{
		cout << "ERROR: Attach camera " << _camera_id
			<< " for image " << _image_id << " failed.\n";
		return false;
		}
	
	_camera->inuse(true);
	
	return true;
}


// AP set correction

DPoint2 Image::ap_set_corr(const DPoint3& pho_c) const
{
	DPoint2 corr(0., 0.);

	vector<APSet*>::const_iterator aps;
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
		{
		if (*aps == 0) continue;
		corr += (*aps)->calc_corr(pho_c);
		}

	return corr;
}


// add subset of derivatives to linear condition equations

void Image::add_a(LinConditionEqn& eqn) const
{
	vector<APSet*>::const_iterator aps;
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
		{
		if (*aps == 0) continue;
		eqn.add_a(&(*aps)->_derivs);
		}
}


// initialize partial derivatives of IO parameters
/*
void Camera::init_derivs(DerivSub& derivs) const
{
	if (_ap_set != 0) _ap_set->init_derivs(derivs);
}
*/


// get partial derivatives of IO parameters

//void Camera::get_derivs(const DPoint2& img_c, DerivSub& derivs) const
void Image::update_derivs(const DPoint3& pho_c)
{
	vector<APSet*>::const_iterator aps;
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
		{
		if (*aps == 0) continue;
		(*aps)->update_derivs(pho_c);
		}
}


// get partial derivatives of IO parameters

void Image::get_derivs(const DPoint3& pho_c, vector<DerivSub>& derivs) const
{
	assert(_ap_sets.size() == derivs.size());

	vector<DerivSub>::iterator drv = derivs.begin();

	vector<APSet*>::const_iterator aps;
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps, ++drv)
		{
		if (*aps == 0)
			{
			cout << "ERROR: Camera::get_derivs: null pointer!" << endl;
			assert(false);
			}
		(*aps)->get_derivs(pho_c, *drv);
		drv->index((*aps)->get_index());
		}
}


// update IO parameter
/*
bool Camera::update_io()
{
	//if (_ap_set == 0) return false;
	//_ap_set->update(_princ_pnt.x, _princ_pnt.y, _focal_len);
	//return true;
	cerr << "ERROR: function failed!\n";
	return false;
}
*/


// image-to-photo-coords. transformation, AP correction included
// ATTENTION: the AP corrections are functions of the photo coords.!!!

void Image::image_to_photo(const DPoint2& img_c, DPoint3& pho_c) const
{
	assert(_camera != 0);
	
	_camera->image_to_photo(img_c, pho_c);

	// approximation only!!!
	DPoint2 corr = ap_set_corr(pho_c);

	pho_c.x -= corr.x;
	pho_c.y -= corr.y;
}


// photo-to-image-coords. transformation, AP correction included

void Image::photo_to_image(const DPoint3& pho_c, DPoint2& img_c) const
{
	assert(_camera != 0);
	
	_camera->photo_to_image(pho_c, img_c);
	
	//img_c += ap_set_corr(pho_c);
}


// sensor/photo coordinates transformation

void Image::sensor_to_photo(const DPoint2& ext_c, DPoint3& pho_c) const
{
	assert(_camera != 0);
	
	DPoint2 img_c;
	_camera->sensor_to_image(ext_c, img_c);
	_camera->image_to_photo(img_c, pho_c);
}


// calculate exterior orientation

bool Image::calc_ext_orientation(DPoint3& pcc, DRotMatrix3& R) const
{
	assert(_station != 0);
	
	const ExtOrientation* eo = dynamic_cast<const ExtOrientation*>(_station);
	assert(eo != 0);
	
	// perspective centre coords.
	eo->get_per_centre(pcc);
	
	// exposure time
	double t = eo->get_time();
	
	// position crrection (optional)
	if (_pos_cor != 0)
		pcc += _pos_cor->calc_corr(t);
	
	// rotation parameters
	size_t n = eo->num_rot_params(); // test missing!
	vector<double> a(n);
	eo->get_rot_params(a);
	
	// attitude correction (optional)
	if (_att_cor != 0)
		{
		DPoint3 da = _att_cor->calc_corr(t);
		a[0] += da.x;
		a[1] += da.y;
		a[2] += da.z;
		}
		
	// rotation matrix
	eo->calc_rot_matrix(R, a);
	//cout << "Image" << endl;
	
	return true;
}


// object-to-photo coords. transformation

bool Image::object_to_photo(const DPoint3& obj_c, DPoint3& pho_c) const
{
	assert(_camera != 0);
	
	DPoint3 pcc;
	DRotMatrix3 R;
	
	if (calc_ext_orientation(pcc, R) != true)
		assert(false);
	
	// set principal distance
	// The z-component of the photo coordinates is initialized with z = -c!
	// The focal length is required for the central projection
	// (collinearity equations).
	// Other image geometries didn't need this information,
	// e.g. DLT. But a (pseudo) camera is still required
	// to define the sensor image coordinate system.
	pho_c.z = -_camera->ck();
	
	// back projection
	return projection(pcc, R, obj_c, pho_c);
}


// object-to-image coords. transformation (UNUSED ARGUMENTS: r)

bool Image::object_to_image(const DPoint3& obj_c, DPoint2& img_c, double r) const
{
	// object-to-photo coords. transformation
	DPoint3 pho_c;
	object_to_photo(obj_c, pho_c);
	
	// photo-to-image coords. transformation, including AP correction
	photo_to_image(pho_c, img_c);
	
	return true;
}


// image-to-object coords. transformation

bool Image::image_to_object(const DPoint2& img_c, DPoint3& obj_c) const
{
	cout << "ERROR: image_to_object() not implemented yet!" << endl;
	assert(false);

	return true;
}


// find corresponding image and object points

int Image::find_connections(const vector<ObjectPoint*>& obpts,
vector<IOPointCon*>& cons) const
{
	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;
		
		// search for corresponding object point
		ObjectPoint* op = DEP::find(obpts, (*ip)->get_id());
		if (op == 0)
			{
			cout << "ATTENTION: Object point " << (*ip)->get_id()
				<< " not found.\n";
			continue;
			}
		//cout << _id << ":" << (*ip)->get_id() << ":\n";

		// might be a control or intersection point
		// but only approximations are required
		if (op->is_approx())
			{
			IOPointCon* c = new IOPointCon;
			assert(c != 0);

			c->ip = *ip;
			c->op = op;

			cons.push_back(c);
			}
		}

	return cons.size();
}


// spatial resection
// Problem: who to handle multiple images to one image geometry
// relation?

bool Image::resection(ostream& os, const vector<ObjectPoint*>& obpts, bool dbg)
{
	assert(true);

	assert(_camera != 0);
	assert(_station != 0);

	// find control points
	//vector<IOPointCon*> copts;
	//int num = find_connections(obpts, copts);
	//cout << "Image::resection: " << num << endl;

	// try resection, method can fail when not enough control points
	///_station->resection(_camera, copts, os, dbg);

	return true;
}


// spatial intersection

bool Image::intersection(const DPoint2& ext_c1, const Image& image2,
const DPoint2& ext_c2, DPoint3& obj_c, double& e) const
{
// 	assert(this->_camera != 0 && this->_station != 0);
// 	assert(image2._camera != 0 && image2._station != 0);
// 	
// 	// sensor-to-image coords. transformation
// 	DPoint2 img_c1, img_c2;
// 	this->_camera->sensor_to_image(ext_c1, img_c1);
// 	image2._camera->sensor_to_image(ext_c2, img_c2);
// 
// 	// image-to-photo coords. transformation
	DPoint3 pho_c1, pho_c2;
// 	this->image_to_photo(img_c1, pho_c1);
// 	image2.image_to_photo(img_c2, pho_c2);
	
	this->sensor_to_photo(ext_c1, pho_c1);
	image2.sensor_to_photo(ext_c2, pho_c2);
	
	// perspective centre coords. and rotation matrix
	DPoint3 pcc1, pcc2;
	DRotMatrix3 rm1, rm2;
	if (this->calc_ext_orientation(pcc1, rm1) != true)
		return false;
	if (image2.calc_ext_orientation(pcc2, rm2) != true)
		return false;
	
	// do spatial intersection
	return DEP::intersection(pcc1, rm1, pho_c1, pcc2, rm2, pho_c2, obj_c, e);
}


// observation equation for intersection

bool Image::observ_eqn_intersec(const string& id, DerivSub& a, DVector& wt,
bool& found) const
{
	ImagePoint* ip = get_image_point(id);
	if (ip == 0)
		{
		found = false;
		return true; // escape, point not found
		}
	found = true;

	DPoint3 pho_c;
	sensor_to_photo(ip->crds, pho_c);
	
	DPoint3 pcc;
	DRotMatrix3 R;
	
	if (calc_ext_orientation(pcc, R) != true)
		{
		cout << "ERROR: Image::observ_eqn_intersec: Zero denominator.\n";
		return false;
		}
		
	double u = R(0,0) * pho_c.x + R(0,1) * pho_c.y + R(0,2) * pho_c.z;
	double v = R(1,0) * pho_c.x + R(1,1) * pho_c.y + R(1,2) * pho_c.z;
	double w = R(2,0) * pho_c.x + R(2,1) * pho_c.y + R(2,2) * pho_c.z;
	if (fabs(w) == 0.)
		{
		cout << "ERROR: Image::observ_eqn_intersec: Zero denominator.\n";
		return false;
		}
		
	double ax = - u / w;
	double ay = - v / w;
	
	double bx = - (ax * pcc.z + pcc.x);
	double by = - (ay * pcc.z + pcc.y);
	
	a(0,0) = 1.;
	a(0,1) = 0.;
	a(0,2) = ax;
	
	a(1,0) = 0.;
	a(1,1) = 1.;
	a(1,2) = ay;
	
	wt(0) = -bx;
	wt(1) = -by;

	return true;
}


// add contribution to normal equations for all image point observations
// using the Gauss-Helmert model

bool Image::normal_eqns(const vector<ObjectPoint*>& obpts,
Matrix<double>& atpa, Vector<double>& atpl, double& ltpl, int& num_obs) const
{
	assert(_station != 0);	// necessary
	assert(_camera != 0);	// necessary
	
	const ExtOrientation* eo = dynamic_cast<const ExtOrientation*>(_station);
	assert(eo != 0);
	
	// Setup sub-matrices of derivatives
	// Always 2 observation equations for each image point.
	// The number of columns might be variable.
	DerivSub	deriv_oc(2,3);		// object points, size always 2x3
	DerivSub	deriv_pcc(2,3);		// per. centre coords., size always 2x3
	DerivSub	deriv_rot(2,eo->num_rot_params());		// rotation params., variable size
	vector<DerivSub> deriv_ap(_camera->num_ap_sets());	// APs, variable size
	DerivSub	deriv_pc;			// position corr., variable size
	DerivSub	deriv_ac;			// attitude corr., variable size
	DerivSub	deriv_obs(2,2);		// observations, size always 2x2

	// derivatives of observations
	deriv_obs.identity();
	
	// set index for EO (valid for the whole image)
	deriv_pcc.index(eo->get_index());
	deriv_rot.index(eo->get_index() + 3);

	// always 2 condition equations for a single imaging ray.
	LinConditionEqn con_eqn;

	// design matrix A
	con_eqn.add_a(&deriv_oc);	// always
	con_eqn.add_a(&deriv_pcc);	// always
	con_eqn.add_a(&deriv_rot);	// always
	con_eqn.add_a(deriv_ap);	// optional
	if (_pos_cor != 0) con_eqn.add_a(&deriv_pc);	//optional
	if (_att_cor != 0) con_eqn.add_a(&deriv_ac);	//optional
	
	// design matrix B
	con_eqn.set_b(deriv_obs);
	//cout << "B: " << deriv_obs << endl;
	
	// perspective centre coords.
	DPoint3 pcc;
	eo->get_per_centre(pcc);
	
	// exposure time
	double t = eo->get_time();
	
	// position correction (optional)
	if (_pos_cor != 0)
		pcc += _pos_cor->calc_corr(t);
	//cout << pcc << endl;
		
	// rotation parameter
	size_t n = eo->num_rot_params();
	vector<double> a(n);
	eo->get_rot_params(a);
		
	// attitude correction (optional)
	if (_att_cor != 0)
		{
		DPoint3 da = _att_cor->calc_corr(t);
		a[0] += da.x;
		a[1] += da.y;
		a[2] += da.z;
		}
	
	// instantenous rotation matrix
	DRotMatrix3 R;
	eo->calc_rot_matrix(R, a);
	//cout << R << endl;
	
	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;
		
		// search for corresponding object point
		ObjectPoint* op = find(obpts, (*ip)->get_id());
		if (op == 0)
			{
			cout << "ATTENTION: Object point " << (*ip)->get_id()
				<< " not found.\n";
			continue;
			}

		//cout << _image_id << ":" << (*ip)->get_id() << ":\n"; cout.flush();

		// set current index of object point
		deriv_oc.index(op->get_index());

		// convert sensor to image std. dev.
		DPoint2 sdev;
		_camera->sdev_sensor_to_image((*ip)->sdev, sdev);
		//cout << (*ip)->get_id() << ": " << sdev << endl;

		// co-factors
		Matrix<double> Qll(2,2);
		Qll(0,0) = sdev.x / AdjObject::lspars.sigma0(); Qll(0,0) *= Qll(0,0);
		Qll(1,1) = sdev.y / AdjObject::lspars.sigma0(); Qll(1,1) *= Qll(1,1);
		con_eqn.set_q(Qll);
		//cout << "Qll: " << Qll << endl;

		// convert sensor into internal image coords.
		DPoint2 img_c;
		_camera->sensor_to_image((*ip)->crds, img_c);
		//cout << "img: " << img_c << endl;

		// convert image into photo coords., AP included
		DPoint3 pho_c;
		image_to_photo(img_c, pho_c);
		//cout << "pho: " << pho_c << endl;

		// partial derivatives of object point coords.
		// and orientation parameters
		// The derivatives of the projection centre coordinates and
		// the object coordinates are always the same only the derivatives
		// of the rotation parameters are different!

		// object coords. reduced to the perspective centre
		DPoint3 d;
		d.x = op->crds.x - pcc.x;
		d.y = op->crds.y - pcc.y;
		d.z = op->crds.z - pcc.z;

		double w = R(0,2) * d.x + R(1,2) * d.y + R(2,2) * d.z;
		if (fabs(w) == 0.)
			{
			cout << "ERROR: Image::normal_eqns: Zero denominator.\n";
			return false;
			}

		double u = R(0,0) * d.x + R(1,0) * d.y + R(2,0) * d.z;
		double v = R(0,1) * d.x + R(1,1) * d.y + R(2,1) * d.z;

		double g = pho_c.z / w;
		double h = g / w;
	
		// derivatives of the object coordinates (always)
		// X-derivatives
		deriv_oc(0,0) = h * (R(0,0) * w - R(0,2) * u);
		deriv_oc(1,0) = h * (R(0,1) * w - R(0,2) * v);

		// Y-derivatives
		deriv_oc(0,1) = h * (R(1,0) * w - R(1,2) * u);
		deriv_oc(1,1) = h * (R(1,1) * w - R(1,2) * v);

		// Z-derivatives
		deriv_oc(0,2) = h * (R(2,0) * w - R(2,2) * u);
		deriv_oc(1,2) = h * (R(2,1) * w - R(2,2) * v);
	
		// The derivatives of the projection centre coordinates are simply
		// the negative derivatives of the object coordinates !
		// Xo-derivatives
		deriv_pcc(0,0) = - deriv_oc(0,0);
		deriv_pcc(1,0) = - deriv_oc(1,0);

		// Yo-derivatives
		deriv_pcc(0,1) = - deriv_oc(0,1);
		deriv_pcc(1,1) = - deriv_oc(1,1);

		// Zo-derivatives
		deriv_pcc(0,2) = - deriv_oc(0,2);
		deriv_pcc(1,2) = - deriv_oc(1,2);
	
		// get derivatives of the rotation parameters
		eo->calc_derivs_rot(deriv_rot, h, u, v, w, d, a);
	
		// derivatives of position correction parameters (optional)
		if (_pos_cor != 0)
			_pos_cor->calc_derivs(t, deriv_pcc, deriv_pc);
		
		// derivatives of attitude correction parameters (optional)
		if (_att_cor != 0)
			_att_cor->calc_derivs(t, deriv_rot, deriv_ac);

		// partial derivatives of APs
		// derivatives are computed with
		switch (AdjObject::lspars.ap_derivs())
			{
			case LSParams::IMAGE_COORDS:
				// image coords. without AP correction!
				_camera->image_to_photo(img_c, pho_c, false);
				break;
				
			case LSParams::COLLINEAR_EQUATION:
				// image coords. derived from collinear equation
				object_to_photo(op->crds, pho_c);
				break;
			
			case LSParams::IMAGE_COORDS_PLUS_APS:
				// image coords. plus AP correction
				// already calculated, see above
				break;
			
			default:
				assert(false);
			}
		_camera->get_derivs(pho_c, deriv_ap);

		// image coords. derived from the unknown parameters
		DPoint2 img_c0;
		//object_to_image(op->crds, img_c0);
		projection(pcc, R, op->crds, pho_c);
		//pho_c.x = pho_c.z * u / w;
		//pho_c.y = pho_c.z * v / w;
		//cout << "pho_c: " << pho_c << endl; cout.flush();
	
		// photo-to-image coords. transformation, including AP correction
		_camera->photo_to_image(pho_c, img_c0);
		//cout << "img_c0: " << img_c0 << endl;
		
		// calculate shorten observations (observed - computed term)
		// misclosure vector
		Vector<double> wt(2);
		wt(0) = img_c.x - img_c0.x;
		wt(1) = img_c.y - img_c0.y;
		con_eqn.set_w(wt);
		//cout << "w: " << wt << endl; cout.flush();
	
		// add contributions to normal equation
		if (!con_eqn.add_contr(atpa, atpl, ltpl))
			return false;
		//con_eqn.print(cout); cout << endl;
		
		num_obs += 2;
		}

	return true;
}


// print residual image coords.

void Image::print_residuals(const vector<ObjectPoint*>& obj_pnts,
DPoint2& sum_v, DPoint2& sum_vv, DPoint2& sum_pvv,
DPoint2& sum_p, int& num, ostream& os) const
{
	assert(_camera != 0);

	Form str8; str8.width(8);
	Form fix10(2); fix10.fixed().width(10);

	DPoint2 img_c0;
	DPoint2 v, vv;
	DPoint2 p;
	double v_abs = 0.;

	// initialize accumulator
	sum_v.x = sum_v.y = 0.;
	sum_vv.x = sum_vv.y = 0.;
	num = 0;
	
    os << "Image ID : " << _image_id << endl << endl;
	os << "  Point ID       vx/um      vy/um     |v|/um\n\n";

	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;

		// find corresponding object point
		ObjectPoint* op = find(obj_pnts, (*ip)->get_id());
		if (op == 0)
			{
			os << "ATTENTION: Object point "
				<< (*ip)->get_id() << " not found" << endl;
			continue;
			}

		// collinear equation -> calculate image coords.
		// Attention: The scanline coordinate is only used by scanner images!!!
		DPoint2 img_c0;
		object_to_image(op->crds, img_c0, (*ip)->crds.y);

		// convert sensor into image coords.
		DPoint2 img_c;
		_camera->sensor_to_image((*ip)->crds, img_c);

		// residuals: Soll - Ist
		v = img_c0 - img_c;
		v_abs = abs(v);
		vv = v * v;

		// convert sensor into image std. dev.
		DPoint2 sdev;
		_camera->sdev_sensor_to_image((*ip)->sdev, sdev);

		// weights
		p.x = AdjObject::lspars.sigma0() / sdev.x;
		p.y = AdjObject::lspars.sigma0() / sdev.y;
		p *= p;
		
		sum_v += v;
		sum_vv += vv;
		sum_pvv += vv * p;
		sum_p += p;
		num++;

		// print residual
		if (AdjObject::lspars.is_small(v.x) || AdjObject::lspars.is_small(v.y))
			{
			os << "  " << str8((*ip)->get_id()) << "  "
				<< fix10(v.x * 1e6) << " " << fix10(v.y * 1e6) << " "
				<< fix10(v_abs * 1e6);

			if (op->is_ctrl())
				{
				os << "   control point for";
				if (op->is_csx()) os << " X";
				if (op->is_csy()) os << " Y";
				if (op->is_csz()) os << " Z";
				}
			os << endl;
			}
		}
	os << endl;

	os << "Number of points       :   " << num << endl;
	os << "RMS of residuals       :   "
		<< fix10(sqrt(sum_vv.x / num) * 1e6) << "   "
		<< fix10(sqrt(sum_vv.y / num) * 1e6) << endl;
	os << "Average of residuals   :   "
		<< fix10(sum_v.x * 1e6 / num) << "   "
		<< fix10(sum_v.y * 1e6 / num) << endl;
	os << endl;
}


// write residuals

void Image::write_residuals(const vector<ObjectPoint*>& obj_pnts,
ostream& os) const
{
	assert(_camera != 0);

	Form str8; str8.width(8);
	Form fix10(3); fix10.fixed().width(10);
	Form fix12(3); fix12.fixed().width(12);

	// scales for sensor coords. and  residuals
	double sc_xy = (_camera->csystem() == Camera::IMAGE) ? (1e6) : (1.);
	const double sc = 1e6; // convert in mue-meter
	DPoint2 img_c0, img_c, v;
	
	os << _image_id << endl;
	os << fix12(_camera->get_size_x() * 1e6) << "  " << fix12(_camera->get_size_y() * 1e6) << endl;

	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;

		// find corresponding object point
		ObjectPoint* op = find(obj_pnts, (*ip)->get_id());
		if (op == 0)
			{
			cout << "ATTENTION: Object point "
				<< (*ip)->get_id() << " not found" << endl;
			continue;
			}

		// collinear equation -> calculate image coords.
		object_to_image(op->crds, img_c0, (*ip)->crds.y);

		// convert sensor into image coords.
		_camera->sensor_to_image((*ip)->crds, img_c);

		// residuals: Soll - Ist
		v = img_c0 - img_c;

		// write ID, image coords. and residuals
		os << "  " << str8((*ip)->get_id()) << "  ";
		os << fix12((*ip)->crds.x * sc_xy) << " " << fix12((*ip)->crds.y * sc_xy);
		os << " " << fix10(v.x * sc) << " " << fix10(v.y * sc) << endl;
		}
	os << "stop-dep\n";
}


// write photo coords.
/*
void Image::write_photo_coords(ostream& os) const
{
	assert(_camera != 0);

	Form str8; str8.width(8);
	Form fix3(3); fix3.fixed().width(14);

	const double sc = 1e6; // convert in mue-meter
	DPoint2 img_c;
	DPoint3 pho_c;

	os << _image_id << endl;

	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;

		_camera->sensor_to_image((*ip)->crds, img_c);
		_camera->image_to_photo(img_c, pho_c);

		//_camera->sdev_sensor_to_image((*ip)->sdev, sdev);

		os << str8((*ip)->get_id())
			<< " " << fix3(pho_c.x * sc) << " " << fix3(pho_c.y * sc)
			<< " " << fix3(pho_c.z * sc) << endl;
		}
	os << "stop-dep\n";
}
*/


// non-member functions ...

} // namespace DEP
