/* $Id: image.h,v 1.16 2007/10/29 11:11:48 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_IMAGE_INCLUDED
#define DEP_IMAGE_INCLUDED

#include "depdef.h"

using namespace GenLib2;

namespace DEP
{
class LSParams;
class DerivSub;
class LinConditionEqn;
class ImagePoint;
class ObjectPoint;
class Camera;
class Station;
class APSet;
class IOPointCon;
class PosCorrection;
class AttCorrection;

/** Image class.
 *  Implementation of the extended photogrammetric model.
 *
 *  Assuming a perspective projection, the mapping from a 3D object coordinate 
 *  system onto the 2D image coordinate system is accomplished via the
 *  collinearity equation. The spatial position and attitude of the camera
 *  station is defined by the Exterior Orientation (EO). 
 *  The spatial position, also called perspective projection centre, is
 *  defined by its coordinates Xo, Yo and Zo. The attitude is defined by
 *  the three rotation angles. In this base class no assumtion is made on 
 *  the parametrisation of the three rotation angles. This spezialisation
 *  is done by the drived classes, e.g. ExtOrientationOPKRot.
 
 *  Also this class ties everything together which belongs to an image:
 *  the image points, the transformation parameters and the imaging camera.
 *
 *  @author Dirk Stallmann
 */
class Image
{
public:
	/// constructor
	Image();

	/// virtual destructor
	virtual ~Image() {}

	/**@name read/print image configuration
	 */
	//@{
		/// read image configuration
		virtual bool read(istream& is, const vector<Camera*>& cams);

		/// print image configuration
		virtual void print(ostream& os) const;
	//@}

	/// get image type 
	virtual string type_str() const;

	/**@name set and get image ID */
	//@{
		/// set image ID
		void get_id(const string& s) { _image_id = s; }

		/// get image ID
		const char* get_id() const { return _image_id.c_str(); }
	//@}
	
	/**@name inuse flag */
	//@{
		/// set inuse flag
		void inuse(bool flag) { _inuse = flag; }

		/// set inuse flag
		void set_inuse() { _inuse = true; }

		/// test inuse flag
		bool is_unused() const { return _inuse == false; }
	//@}

	/**@name image points
	 *  Coordinates of the image points in the sensor coordinate system.
	 *  The attached camera defines the sensor coordinate system!
	 */
	//@{
		/// image points
		vector<ImagePoint*> _image_points;

		/// read image coords.
		bool read_image_coords(istream& is, bool sdev_flg, const DPoint2& sdev);

		/// write image coords.
		void write_image_coords(ostream& is) const;
	
		/// print image coords.
		void print_image_coords(ostream& os) const;

		/// append or replace image point
		bool add_image_point(ImagePoint* ip);

		/// count number of image points
		int count_image_points() const;

		/// check if image point exist
		bool image_point_exist(const string& id) const;

		/// find image point by ID
		int find_image_point(const string& id) const;

		/// get image point by ID
		ImagePoint* get_image_point(const string& id) const;

		/// remove image point
		bool remove_image_point(const string& id);
			
		// write adjusted image coords.
		//void write_adjust_image_coords(ostream& os) const;
	//@}

	/**@name attach mandatory objects */
	//@{
		/// attach station
		virtual bool attach(const vector<Station*>& stats);

		/// attach camera
		bool attach(const vector<Camera*>& cams);
		
		// attach AP set
		//void attach(APSet* aps) { _ap_sets.push_back(aps); }
	//@}
	
	/**@name attach optional objects */
	//@{
		/// attach position correction
		void attach(PosCorrection* p)
		{
			assert(_pos_cor == 0);
			_pos_cor = p;
		}
		
		/// attach attitude correction
		void attach(AttCorrection* p)
		{
			assert(_att_cor == 0);
			_att_cor = p;
		}
	//@}
	
	/* AP related functions
	 *  But this class has a vector of pointers to Additional Parameters
	 *  (AP) sets.
	 */
	//@{
		// get AP set correction
		DPoint2 ap_set_corr(const DPoint3& pho_c) const;
		
		// add subset of derivatives to linear condition equations
		void add_a(LinConditionEqn& eqn) const;

		// get partial derivatives of IO parameters
		void update_derivs(const DPoint3& pho_c);
		//void get_derivs(const DPoint2& img_c, vector<DerivSub>& derivs) const;

		// get partial derivatives of IO parameters
		void get_derivs(const DPoint3& pho_c, vector<DerivSub>& derivs) const;
	//@}
	
	/**@name image/photo coordinates transformation
	 *  To convert between image and photo coordinates.
	 *  Photo coordinates are image coordinates reduced to the principal
	 *  point. Corrections of APs are included!
	 */
	//@{
		/// image-to-photo coordinates transformation
		void image_to_photo(const DPoint2& img_c, DPoint3& pho_c) const;

		/// photo-to-image-coords. transformation
		void photo_to_image(const DPoint3& pho_c, DPoint2& img_c) const;
	//@}
	
	/// sensor-to-photo coordinates transformation
	void sensor_to_photo(const DPoint2& ext_c, DPoint3& pho_c) const;

	/// calculate exterior orientation
	virtual bool calc_ext_orientation(DPoint3& pcc, DRotMatrix3& rm) const;
	
	/// object-to-photo-coords. transformation
	virtual bool object_to_photo(const DPoint3& obj_c, DPoint3& pho_c) const;
	
	/**@name coordinate transformation */
	//@{
		/**@name object-to-image coords. transformation
		 *
		 *  @param obj_c (IN) object coordinates
		 *  @param img_c (OUT) image coordinates
		 *  @param r (IN) row/scanline coordinate (unsed by frame images!)
		 *  @return true = no problem, false = error encountered 
		 */
		virtual bool object_to_image(const DPoint3& obj_c,
			DPoint2& img_c, double r=0) const;

		/// image-to-object coords. transformation (ray-Z-plane intersection)
		virtual bool image_to_object(const DPoint2& img_c,
			DPoint3& obj_c) const;
	//@}
	
	/** Find corresponding image and object points.
	 *  Loop over all image points and find corresponding object point.
	 *
	 *  @param obpts vector with object points
	 *  @param cons vector with with corresponding image and object points
	 *  @return number of connections
	 */
	int find_connections(const vector<ObjectPoint*>& obpts,
		vector<IOPointCon*>& cons) const;

	/// spatial resection
	virtual bool resection(ostream& os, const vector<ObjectPoint*>& obpts, bool dbg);
	
	/// spatial intersection
	virtual bool intersection(const DPoint2& ext_c1, const Image& image2,
		const DPoint2& ext_c2, DPoint3& obj_c, double& e) const;
	
	/** observation equation for intersection
	 *  See Schmid, BuL 1958, Nr 4, Seite 112-113.
	 *
	 *  @param id (IN) object points ID
	 *  @param a (OUT) coeffs. for design matrix A
	 *  @param wt (OUT) misclosures
	 *  @param found (OUT) true = point found, false = point not found
	 *  @return true = no problem, false = error encountered
	 */
	virtual bool observ_eqn_intersec(const string& id, DerivSub& a, DVector& wt,
		bool& found) const;

	/// add constribution to normal equations for all image point observations
	virtual bool normal_eqns(const vector<ObjectPoint*>& obpts,
		Matrix<double>& atpa, Vector<double>& atpl, double& ltpl,
		int& num_obs) const;
	
	/// print residuals
	virtual void print_residuals(const vector<ObjectPoint*>& obj_pnts,
		DPoint2& sum_v, DPoint2& sum_vv, DPoint2& sum_pvv,
		DPoint2& sum_p, int& num, ostream& os) const;

	/// write residuals (in mue-meter)
	virtual void write_residuals(const vector<ObjectPoint*>& obj_pnts,
		ostream& os) const;

	// write photo coords. (in mue-meter)
	//void write_photo_coords(ostream& os) const;

protected:
	/// image ID
	string		_image_id;
	
	/// station ID
	string		_station_id;
	
	/// station (mandatory)
	Station*	_station;
	
	/// camera ID
	string		_camera_id;
	
	/// imaging camera (mandatory)
	Camera*		_camera;
	
	/* AP set list
	 *  The AP sets modelling systematic errors in image space.
	 *  The option to connect the image with many AP sets gives
	 *  the possibility to combine different AP sets to model
	 *  various sources of systematic errors.
	 *  So the number of unknown parameters can vary with the type
	 *  and the number of connected AP sets.
	 */
	vector<APSet*>		_ap_sets;
	
	/// inuse flag
	bool		_inuse;
	
	/// position correction (optional)
	PosCorrection*		_pos_cor;
	
	/// attitude correction (optional)
	AttCorrection*		_att_cor;
};


// non-member functions ...

} // namespace DEP

#endif
