/* $Id: imggeo.cpp,v 1.12 2007/12/10 06:57:45 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "imggeo.h"
//#include "cholesky.h"
//#include "camera.h"
//#include "iopntcon.h"
//#include "lobseqn.h"

namespace DEP
{

// allocate memory for parameters and std. dev.

void TransfParams::allocate(size_t sz)
{
	_size = sz;
	_a = new double[_size];
	assert(_a != 0);
	_sdev_a = new double[_size];
	assert(_sdev_a != 0);
}


// free memory for parameters and std. dev.

void TransfParams::free()
{
	delete[] _a;
	delete[] _sdev_a;
}


// new size/resize for parameters and std. dev.

void TransfParams::newsize(size_t sz)
{
	if (_size == sz) return;
	free();
	allocate(sz);
}


// constructor

TransfParams::TransfParams(size_t sz) : Station(sz)
{
	allocate(sz);
}


// destructor

TransfParams::~TransfParams()
{
	free();
}


// read

bool TransfParams::read(istream& is)
{
	// station ID
	is >> _id;
	if (!is)
		{
		cout << "ERROR: Reading station ID failed.\n";
		return false;
		}

	// read values and std. devs.
	for (size_t i = 0; i < _size; i++)
		{
		is >> _a[i] >> _sdev_a[i];
		if (!is)
			{
			cout << "ERROR: Reading values and std. devs. failed.\n";
			return false;
			}
		}

	return true;
}


// write

void TransfParams::write(ostream& os) const
{
	Form sci6(6); sci6.scientific().width(12);
	Form sci15(15); sci15.scientific().width(23);

	// type
	os << key_str() << "   ";

	// ID
	os << _id << endl;

	// parameters ans std. devs.
	for (size_t i = 0; i < _size; i++)
		os << sci15(_a[i]) << "   " << sci6(_sdev_a[i]) << endl;
	os << endl;
}


// print

void TransfParams::print(ostream& os) const
{
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(6); fix.fixed().width(12);
	Form sci(5); sci.scientific().width(12);

	os << "Station ID    :  " << _id << endl;
	os << "Type          :  " << type_str() << endl;
	os << endl;

	os << "  Param.     State         Value       Std.dev.\n\n";
	for (size_t i = 0; i < _size; i++)
		os << "  " << str8(uname(i)) << "   " << str4(state_str(_sdev_a[i]))
			<< "   " << sci(_a[i]) << "   " << sci(_sdev_a[i]) << endl;

	os << endl;
}


// get parameter name

string TransfParams::uname(size_t i) const
{
	assert(i < _size);

	ostringstream oss;
	oss << "a" << (i + 1);

	return oss.str();
}


// set upari

bool TransfParams::set_upari()
{
	if (create() != true) return false;

	for (size_t i = 0; i < _size; i++)
		set(i, &_a[i], &_sdev_a[i]);

	return true;
}


// spatial resection
/*
bool TransfParams::resection(Camera* cam,
const vector<IOPointCon*>& copts, ostream* os, bool dbg)
{
	if (*os)
		{
		(*os) << "Station ID    :  " << _id << endl;
		(*os) << endl;
		}

	int n = 0;		// number of observations

	// estimate solution by adjustment
	DMatrix atpa(_size, _size, 0.);
	DVector atpl(_size);
	DVector dec_vec(_size);
	DVector x(_size);
	double ltpl = 0.;

	// setup sub-matrices of derivatives
	DerivSub	deriv_oc(2,3);		// required, but UNUSED
	DerivSub	deriv_eo;			// EO
	//DerivSub	deriv_pc;			// UNUSED
	//DerivSub	deriv_ac;			// UNUSED
	DerivSub	deriv_obs(2,2);		// UNUSED

	// set index for EO (valid for the whole image)
	deriv_eo.index(0);

	LinObsEqn obs_eqn(2);

	// design matrix A
	obs_eqn.add_a(&deriv_eo);

	if (*os) (*os) << "Using control points:\n";

	// loop over all control points
	vector<IOPointCon*>::const_iterator cp;
	for (cp = copts.begin(); cp != copts.end(); ++cp)
		{
		ImagePoint* ip = (*cp)->ip;
		assert(ip != 0);

		ObjectPoint* op = (*cp)->op;
		assert(op != 0);

		if (*os) (*os) << "  " << ip->get_id() << endl;

		// check on approximations,
		// use either control and intersection points
		// with approximations for resection
		if (!op->is_approx()) continue;

		//cout << "obj_c: " << op->crds << endl;
		//cout << "ext_c: " << ip->crds << endl;

		// image to photo coords, the camera constant is ignored!
		DPoint2 img_c;
		cam->sensor_to_image(ip->crds, img_c);
		//cout << "img_c: " << img_c << endl;

		DPoint3 pho_c;
		cam->image_to_photo(img_c, pho_c);
		//cout << "pho_c: " << pho_c << endl;

		// partial derivatives
		//get_derivs(pho_c, op->crds, deriv_oc, deriv_eo,
			//deriv_pc, deriv_ac, deriv_obs);
		get_derivs(pho_c, op->crds, deriv_eo);

		// convert external to image std. dev.
		DPoint2 sdev;
		cam->sdev_sensor_to_image(ip->sdev, sdev);

		// weights
		double px = sdev.x / AdjObject::lspars.sigma0(); px *= px;
		double py = sdev.y / AdjObject::lspars.sigma0(); py *= py;
		obs_eqn.set_p(0, px);
		obs_eqn.set_p(1, py);
		//cout << "px= " << px << endl;
		//cout << "py= " << py << endl;

		// image coords. derived from the unknown parameters
		//DPoint img_c0;
		//object_to_image(op->crds, img_c0);
		//cout << "img0: " << img_c0 << endl;

		// calculate shorten observations
		double lx = pho_c.x;
		double ly = pho_c.y;
		obs_eqn.set_l(0, lx);
		obs_eqn.set_l(1, ly);
		//cout << "lx= " << lx << endl;
		//cout << "ly= " << ly << endl;

		// add contributions to normal equation
		if (!obs_eqn.add_contr(atpa, atpl, ltpl))
			{
			cout << "TransfParams::resection: "
				"Add contributions to normal equation failed.\n";
			return false;
			}

		n++;
		}

	// use stochatic weighting to eliminate unknowns
	//_index = 0;		// set local index
	//cout << _index << endl;
	//cout << _size << endl;
	stoch_weighting(atpa, atpl, ltpl);

	// decomposition
	if (chol_dec(atpa, dec_vec) != 0)
		{
		cout << "TransfParams::resection: "
			"Solving of normal equations failed.\n";
		return false;
		}

	// backsubstitution
	chol_sol(atpa, dec_vec, atpl, x);
	//cout << x << endl;

	// redundancy
	int u = _size;
	int f = 2 * n - u;
	if (f < 0)
		{
		cout << "TransfParams::resection: "
			"Insufficient ground control support.\n";
		return false;
		}

	// vtpv
	double vtpv = ltpl - dot_prod(x, atpl);
	//cout << "vtpv= " << vtpv << endl;

	// sigma node
	double s0 = sqrt(vtpv / double(f));
	//cout << "s0= " << s0 << endl;

	Form sci2(2); sci2.scientific();
	if (*os)
		{
		cout << endl;
		(*os) << "Adjustment results\n";
		(*os) << "------------------\n";
		(*os) << "Number of observations  :  " << n << endl;
		(*os) << "Number of unknowns      :  " << u << endl;
		(*os) << "Redundancy              :  " << f << endl;
		(*os) << "vtpv                    :  " << sci2(vtpv) << endl;
		(*os) << "Sigma 0 a posteriori    :  ";
		if (s0 == 0.)
			cout << "undefined, no redundancy!\n";
		else
			cout << sci2(s0) << endl;
		cout << endl;
		}

	// store geometry parameter and new initial value
	for (size_t i = 0; i < _size; i++)
		_a[i] = _upari[i].ival = x(i);
	_approx_flag = true;

	// print approximations
	if (*os)
		{
		(*os) << "Approximations\n";
		(*os) << "--------------\n\n";
		}
	print(*os);
	cout << endl;

	return true;
}*/


// non-member functions ...

} // namespace DEP
