/**********************************************************************/
//
// FILE:    imuquat.cpp
//          IMU misalignment class functions.
// CREATED: 01 Nov 04  (ds)
// CHANGES: see include file
// REFS:    dito 
//
// (c) 2004 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#include "imuquat.h"
#include "lobseqn.h"

namespace DEP
{

// get type name

string IMUMisalignmentQuaternion::type_str() const
{
	ostringstream oss;
	oss << "unit quaternion";
	return oss.str();
}


// get key name

string IMUMisalignmentQuaternion::key_str() const
{
	ostringstream oss;
	oss << "unit-quaternion";
	return oss.str();
}


// get parameter name

string IMUMisalignmentQuaternion::uname(size_t i) const
{
	ostringstream oss;
	
	switch (i)
		{
		case 0: oss << "q0^b"; break;
		case 1: oss << "q1^b"; break;
		case 2: oss << "q2^b"; break;
		case 3: oss << "q3^b"; break;
		default:
			assert(false);
			break;
		}
	
	return oss.str();
}


// add weighted constraints

void IMUMisalignmentQuaternion::add_constraints(Matrix<double>& atpa,
Vector<double>& atpl, double& ltpl) const
{
	// one observation equation
	LinObsEqn eqn(1);

	// setup sub-matrices of derivatives
	DerivSub deriv(1,_size);

	// set index
	deriv.index(_index);

	// design matrix A
	deriv(0,0) = 2. * _a[0];
	deriv(0,1) = 2. * _a[1];
	deriv(0,2) = 2. * _a[2];
	deriv(0,3) = 2. * _a[3];
	eqn.add_a(&deriv);

	// weight
	eqn.set_p(0, lspars.constr_wt());

	// misclosure vector
	double l = 1. - _a[0] * _a[0] - _a[1] * _a[1] - _a[2] * _a[2]
		- _a[3] * _a[3];
	eqn.set_l(0, l);

	// add contributions to normal equation
	eqn.add_contr(atpa, atpl, ltpl);
}


// calculate derivatives

void IMUMisalignmentQuaternion::calc_derivs(DerivSub& deriv,
double h, double u, double v, double w,
const DPoint3& d, const DRotMatrix3& R) const
{
	deriv.index(_index);
	deriv.newsize(2, 4);
	
	DRotMatrix3 dM_m0;
	DRotMatrix3 dM_m1;
	DRotMatrix3 dM_m2;
	DRotMatrix3 dM_m3;
	dm(dM_m0, dM_m1, dM_m2, dM_m3);
	
	DRotMatrix3 dN_m0 = R * dM_m0;
	DRotMatrix3 dN_m1 = R * dM_m1;
	DRotMatrix3 dN_m2 = R * dM_m2;
	DRotMatrix3 dN_m3 = R * dM_m3;
	
	DPoint3 dd_m0; mult_inv(dd_m0, dN_m0, d);
	DPoint3 dd_m1; mult_inv(dd_m1, dN_m1, d);
	DPoint3 dd_m2; mult_inv(dd_m2, dN_m2, d);
	DPoint3 dd_m3; mult_inv(dd_m3, dN_m3, d);
	
	deriv(0,0) = h * (w * dd_m0.x - u * dd_m0.z);
	deriv(1,0) = h * (w * dd_m0.y - v * dd_m0.z);
	
	deriv(0,1) = h * (w * dd_m1.x - u * dd_m1.z);
	deriv(1,1) = h * (w * dd_m1.y - v * dd_m1.z);
	
	deriv(0,2) = h * (w * dd_m2.x - u * dd_m2.z);
	deriv(1,2) = h * (w * dd_m2.y - v * dd_m2.z);
	
	deriv(0,3) = h * (w * dd_m3.x - u * dd_m3.z);
	deriv(1,3) = h * (w * dd_m3.y - v * dd_m3.z);
}

} // namespace DEP
